"""Tests for the `ships` module"""

import os
import pandas as pd
import xarray as xr
import datetime as dt
import tropycal.tracks as tracks

def read_ships():

    #Read sample subset of HURDATv2 dataset, and retrieve sample storm data
    test_data_dir = os.path.join(os.path.dirname(__file__), '../data')
    filepath = os.path.join(test_data_dir, 'sample_hurdat.txt')
    basin = tracks.TrackDataset('north_atlantic', atlantic_url=filepath)

    #Retrieve Hurricane Ida SHIPS data
    storm = basin.get_storm(('ida',2021))
    return storm.get_ships(dt.datetime(2021, 8, 27, 0))

#==========================================================================================
# Test functions not involving internet connection
#==========================================================================================

def test_get_ships():
    """Tests fetching SHIPS data"""
    
    #Retrieve data for Hurricane Ida (2021)
    ships = read_ships()
    
    #Check attributes
    expected_output = {
        'storm_bearing_deg': 320,
        'storm_motion_kt': 10,
        'max_wind_t-12_kt': 30,
        'steering_level_pres_hpa': 500,
        'steering_level_pres_mean_hpa': 620,
        'brightness_temp_stdev': 21.5,
        'brightness_temp_stdev_mean': 14.5,
        'pixels_below_-20c': 60.0,
        'pixels_below_-20c_mean': 65.0,
        'lat': 18.0,
        'lon': -80.1,
        'storm_name': 'IDA',
        'forecast_init': dt.datetime(2021, 8, 27, 0, 0),
    }
    assert ships.attrs == expected_output

def test_get_snapshot():
    """Tests retrieving Ships data valid at a single time"""
    
    ships = read_ships()
    
    output = ships.get_snapshot(hour=48)
    
    expected_output = {
        'fhr': 48,
        'vmax_noland_kt': 65,
        'vmax_land_kt': 64,
        'vmax_lgem_kt': 59,
        'storm_type': 'TROP',
        'shear_kt': 6,
        'shear_adj_kt': 2,
        'shear_dir': 221,
        'sst_c': 30.1,
        'vmax_pot_kt': 172,
        '200mb_temp_c': -52.0,
        'thetae_dev_c': 7,
        '700_500_rh': 69,
        'model_vortex_kt': 21,
        '850mb_env_vort': 2,
        '200mb_div': 1,
        '700_850_tadv': 10,
        'dist_land_km': 378,
        'lat': 25.8,
        'lon': -88.0,
        'storm_speed_kt': 12,
        'heat_content': 128
    }
    assert output == expected_output

def test_get_ri_prob():
    """Tests fetching RI probability"""
    
    ships = read_ships()
    
    output = ships.get_ri_prob()
    
    expected_output = {
        '20kt/12hr': {'probability': 5, 'climo_mean': 4.9, 'prob / climo': 1.0},
        '25kt/24hr': {'probability': 16, 'climo_mean': 10.9, 'prob / climo': 1.5},
        '30kt/24hr': {'probability': 10, 'climo_mean': 6.8, 'prob / climo': 1.4},
        '35kt/24hr': {'probability': 8, 'climo_mean': 3.9, 'prob / climo': 2.2},
        '40kt/24hr': {'probability': 6, 'climo_mean': 2.4, 'prob / climo': 2.6},
        '45kt/36hr': {'probability': 11, 'climo_mean': 4.6, 'prob / climo': 2.3},
        '55kt/48hr': {'probability': 20, 'climo_mean': 4.7, 'prob / climo': 4.3},
        '65kt/72hr': {'probability': 44, 'climo_mean': 5.3, 'prob / climo': 8.4}
    }
    assert output == expected_output

def test_to_dataframe():
    """Test object can be converted to a pandas DataFrame"""
    
    ships = read_ships()
    
    assert isinstance(ships.to_dataframe(),pd.DataFrame) is True

def test_to_xarray():
    """Test object can be converted to an xarray Dataset"""
    
    ships = read_ships()
    
    assert isinstance(ships.to_xarray(),xr.Dataset) is True



import warnings
from datetime import datetime as dt

from ..plot import Plot

# Import tools
from .tools import *
from ..utils import *

try:
    from cartopy import crs as ccrs
except:
    warnings.warn(
        "Warning: Cartopy is not installed in your python environment. Plotting functions will not work.")

try:
    import matplotlib.lines as mlines
    import matplotlib.patheffects as path_effects
    import matplotlib.pyplot as plt
    import matplotlib.patches as mpatches

except:
    warnings.warn(
        "Warning: Matplotlib is not installed in your python environment. Plotting functions will not work.")


class TornadoPlot(Plot):

    def __init__(self):

        self.use_credit = True

    def plot_tornadoes(self, tornado, domain="east_conus", plotPPH=False, ax=None, return_ax=False, return_domain=False, prop={}, map_prop={}):
        r"""
        Creates a plot of a single storm track.

        Parameters
        ----------
        storm : str, tuple or dict
            Requested storm. Can be either string of storm ID (e.g., "AL052019"), tuple with storm name and year (e.g., ("Matthew",2016)), or a dict entry.
        domain : str
            Domain for the plot. Can be one of the following:

            * **dynamic** - default. Dynamically focuses the domain using the tornado track(s) plotted.
            * **north_atlantic** - North Atlantic Ocean basin.
            * **conus** - Contiguous United States.
            * **east_conus** - Eastern Contiguous United States and western Atlantic.
            * **lonW/lonE/latS/latN** - Custom plot domain.
        ax : axes
            Instance of axes to plot on. If none, one will be generated. Default is none.
        return_ax : bool
            Whether to return axis at the end of the function. If false, plot will be displayed on the screen. Default is false.
        prop : dict
            Property of storm track lines.
        map_prop : dict
            Property of cartopy map.
        """

        # Set default properties
        default_prop = {'plotType': 'tracks', 'PPHcolors': 'spc', 'PPHlevels': [2, 5, 10, 15, 30, 45, 60, 100],
                        'EFcolors': 'default', 'linewidth': 1.5, 'ms': 7.5}
        default_map_prop = {'res': 'm', 'land_color': '#FBF5EA', 'ocean_color': '#EDFBFF',
                            'linewidth': 0.5, 'linecolor': 'k', 'figsize': (14, 9), 'dpi': 200}

        # Initialize plot
        prop = self.add_prop(prop, default_prop)
        map_prop = self.add_prop(map_prop, default_map_prop)
        self.plot_init(ax, map_prop)

        # set default properties
        input_prop = prop
        input_map_prop = map_prop

        # --------------------------------------------------------------------------------------

        # Keep record of lat/lon coordinate extrema
        max_lat = None
        min_lat = None
        max_lon = None
        min_lon = None

        # Check for storm type, then get data for storm
        try:
            tornado_data = tornado
        except:
            raise RuntimeError("Error: tornado must be dataframe")

        # Retrieve storm data
        slat = tornado_data['slat']
        slon = tornado_data['slon']
        elat = tornado_data['elat']
        elon = tornado_data['elon']
        mag = tornado_data['mag']

        mnlat = (slat+elat)*.5
        mnlon = (slon+elon)*.5

        # Add to coordinate extrema
        if max_lat is None:
            max_lat = max(mnlat)
        else:
            if max(mnlat) > max_lat:
                max_lat = max(mnlat)
        if min_lat is None:
            min_lat = min(mnlat)
        else:
            if min(mnlat) < min_lat:
                min_lat = min(mnlat)
        if max_lon is None:
            max_lon = max(mnlon)
        else:
            if max(mnlon) > max_lon:
                max_lon = max(mnlon)
        if min_lon is None:
            min_lon = min(mnlon)
        else:
            if min(mnlon) < min_lon:
                min_lon = min(mnlon)

        # Plot PPH
        if plotPPH in ['total', 'daily', True]:
            if plotPPH == True:
                plotPPH = 'daily'
            PPH, longrid, latgrid = getPPH(tornado_data, method=plotPPH)

            colors, clevs = get_colors_pph(
                plotPPH, prop['PPHcolors'], prop['PPHlevels'])

            cbmap = self.ax.contourf(longrid, latgrid, PPH,
                                     levels=clevs, colors=colors, alpha=0.5)

        # Plot tornadoes as specified
        EFcolors = get_colors_ef(prop['EFcolors'])

        tornado_data = tornado_data.sort_values('mag')
        for _, row in tornado_data.iterrows():
            plt.plot([row['slon'], row['elon']+.01], [row['slat'], row['elat']+.01],
                     lw=prop['linewidth'], color=EFcolors[row['mag']],
                     path_effects=[path_effects.Stroke(
                         linewidth=prop['linewidth']*1.5, foreground='w'), path_effects.Normal()],
                     transform=ccrs.PlateCarree())

        # --------------------------------------------------------------------------------------

        # Storm-centered plot domain
        if domain == "dynamic":

            bound_w, bound_e, bound_s, bound_n = self.dynamic_map_extent(
                min_lon, max_lon, min_lat, max_lat)
            self.ax.set_extent(
                [bound_w, bound_e, bound_s, bound_n], crs=ccrs.PlateCarree())

        # Pre-generated or custom domain
        else:
            bound_w, bound_e, bound_s, bound_n = self.set_projection(domain)

        # Determine number of lat/lon lines to use for parallels & meridians
        self.plot_lat_lon_lines([bound_w, bound_e, bound_s, bound_n])

        # --------------------------------------------------------------------------------------

        # Add left title
        PPH_title = ''
        if plotPPH in ['total', True]:
            PPH_title = ' and total PPH (%)'
        if plotPPH == 'daily':
            PPH_title = ' and daily PPH (%)'
        self.ax.set_title('Tornado tracks'+PPH_title,
                          loc='left', fontsize=17, fontweight='bold')

        # Add right title
        # max_PPH = max(PPH)
        start_time = dt.strftime(
            min(tornado_data['UTC_time']), '%H:%M UTC %d %b %Y')
        end_time = dt.strftime(
            max(tornado_data['UTC_time']), '%H:%M UTC %d %b %Y')
        self.ax.set_title(
            f'Start ... {start_time}\nEnd ... {end_time}', loc='right', fontsize=13)

        # --------------------------------------------------------------------------------------

        # Add legend
        handles = []
        for ef, color in enumerate(EFcolors):
            count = len(tornado_data[tornado_data['mag'] == ef])
            handles.append(mlines.Line2D([], [], linestyle='-',
                           color=color, label=f'EF-{ef} ({count})'))
        leg_tor = self.ax.legend(
            handles=handles, loc='lower left', fancybox=True, framealpha=0, fontsize=11.5)
        leg_tor.set_zorder(101)
        plt.draw()

        # Get the bbox
        try:
            bb = leg_tor.legendPatch.get_bbox().inverse_transformed(self.fig.transFigure)
        except:
            bb = leg_tor.legendPatch.get_bbox().transformed(self.fig.transFigure.inverted())
        bb_ax = self.ax.get_position()

        rectangle = mpatches.Rectangle((bb_ax.x0, bb_ax.y0), bb.width+bb.x0-bb_ax.x0, bb.height+2*bb.y0-2*bb_ax.y0,
                                       fc='w', edgecolor='0.8', alpha=0.8,
                                       transform=self.fig.transFigure, zorder=100)

        # Add PPH colorbar
        if plotPPH != False:

            # Define colorbar axis
            cax = self.fig.add_axes(
                [bb.width + 3*bb.x0 - 2*bb_ax.x0, bb.y0, 0.015, bb.height])
            cbar = self.fig.colorbar(cbmap, cax=cax, orientation='vertical')
            iticks = round(len(clevs)/len(cbar.ax.get_yticks()))
            cbar.ax.set_yticklabels([round(clevs[i], 1) for i in range(
                0, len(clevs), iticks)], fontsize=11.5, color='k')

            rectangle = mpatches.Rectangle((bb_ax.x0, bb_ax.y0), bb.width+bb.x0-bb_ax.x0+.06, bb.height+2*bb.y0-2*bb_ax.y0,
                                           fc='w', edgecolor='0.8', alpha=0.8,
                                           transform=self.fig.transFigure, zorder=100)

        self.ax.add_patch(rectangle)

        # add credit
        text = self.plot_credit()
        self.add_credit(text)

        # Return axis if specified, otherwise display figure
        if ax is not None or return_ax:
            if return_domain:
                return self.ax, leg_tor, {'n': bound_n, 'e': bound_e, 's': bound_s, 'w': bound_w}
            else:
                return self.ax, leg_tor
        else:
            plt.show()
            plt.close()



r"""Functionality for storing and analyzing a year/season of cyclones."""

import numpy as np

# Import internal scripts
from .plot import TrackPlot
from .storm import Storm

# Import tools
from .tools import *
from ..utils import *
from .. import constants


class Season:

    r"""
    Initializes an instance of Season, retrieved via ``TrackDataset.get_season()``.

    Parameters
    ----------
    season : dict
        Dict entry containing all storms within the requested season.
    info : dict
        Dict entry containing general information about the season.

    Returns
    -------
    Season
        Instance of a Season object.
    """

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __add__(self, new):
        # Add seasons

        # Ensure data sources and basins are the same
        if self.source_basin != new.source_basin:
            msg = 'Seasons can only be added for the same basin.'
            raise ValueError(msg)
        if self.source != new.source:
            msg = 'Seasons can only be added from the same source.'
            raise ValueError(msg)

        # Retrieve old & new dict entries
        dict_original = self.dict.copy()
        dict_new = new.dict.copy()

        # Retrieve copy of coordinates
        new_attrs = self.attrs.copy()

        # Add year to list of years
        if isinstance(self.attrs['year'], int):
            new_attrs['year'] = [self.year, new.year]
        else:
            new_attrs['year'].append(new.year)

        # Sort list of years
        new_attrs['year'] = (np.sort(new_attrs['year'])).tolist()

        # Update dict
        dict_original.update(dict_new)

        # Iterate over every year to create a new dict
        new_dict = {}
        for year in new_attrs['year']:
            for key in dict_original.keys():
                if dict_original[key]['season'] == year:
                    new_dict[key] = dict_original[key]

        # Return new Season object
        return Season(new_dict, new_attrs)

    def __repr__(self):

        # Label object
        summary = ["<tropycal.tracks.Season>"]

        # Format keys for summary
        season_summary = self.summary()
        summary_keys = {
            'Total Storms': season_summary['season_storms'],
            'Named Storms': season_summary['season_named'],
            'Hurricanes': season_summary['season_hurricane'],
            'Major Hurricanes': season_summary['season_major'],
            'Season ACE': season_summary['season_ace'],
        }

        # Add season summary
        summary.append("Season Summary:")
        add_space = np.max([len(key) for key in summary_keys.keys()]) + 3
        for key in summary_keys.keys():
            key_name = key + ":"
            summary.append(
                f'{" "*4}{key_name:<{add_space}}{summary_keys[key]}')

        # Add additional information
        summary.append("\nMore Information:")
        add_space = np.max([len(key) for key in self.attrs.keys()]) + 3
        for key in self.attrs.keys():
            key_name = key + ":"
            summary.append(f'{" "*4}{key_name:<{add_space}}{self.attrs[key]}')

        return "\n".join(summary)

    def __init__(self, season, info):

        # Save the dict entry of the season
        self.dict = season

        # Add other attributes about the storm
        keys = info.keys()
        self.attrs = {}
        for key in keys:
            if not isinstance(info[key], list) and not isinstance(info[key], dict):
                self[key] = info[key]
                self.attrs[key] = info[key]
            if isinstance(info[key], list) and key == 'year':
                self[key] = info[key]
                self.attrs[key] = info[key]

    def to_dataframe(self):
        r"""
        Converts the season dict into a pandas DataFrame object.

        Returns
        -------
        `pandas.DataFrame`
            A pandas DataFrame object containing information about the season.
        """

        # Try importing pandas
        try:
            import pandas as pd
        except ImportError as e:
            raise RuntimeError(
                "Error: pandas is not available. Install pandas in order to use this function.") from e

        # Get season info
        season_info = self.summary()
        season_info_keys = season_info['id']

        # Set up empty dict for dataframe
        ds = {
            'id': [],
            'name': [],
            'vmax': [],
            'mslp': [],
            'category': [],
            'ace': [],
            'start_time': [],
            'end_time': [],
            'start_lat': [],
            'start_lon': [],
        }

        # Add every key containing a list into the dict
        keys = [k for k in self.dict.keys()]
        for key in keys:
            # Get tropical duration
            temp_type = np.array(self.dict[key]['type'])
            tropical_idx = np.where((temp_type == 'SS') | (temp_type == 'SD') | (temp_type == 'TD') | (
                temp_type == 'TS') | (temp_type == 'HU') | (temp_type == 'TY') | (temp_type == 'ST'))
            if key in season_info_keys:
                sidx = season_info_keys.index(key)
                ds['id'].append(key)
                ds['name'].append(self.dict[key]['name'])
                ds['vmax'].append(season_info['max_wspd'][sidx])
                ds['mslp'].append(season_info['min_mslp'][sidx])
                ds['category'].append(season_info['category'][sidx])
                ds['start_time'].append(
                    np.array(self.dict[key]['time'])[tropical_idx][0])
                ds['end_time'].append(
                    np.array(self.dict[key]['time'])[tropical_idx][-1])
                ds['start_lat'].append(
                    np.array(self.dict[key]['lat'])[tropical_idx][0])
                ds['start_lon'].append(
                    np.array(self.dict[key]['lon'])[tropical_idx][0])
                ds['ace'].append(np.round(season_info['ace'][sidx], 1))

        # Convert entire dict to a DataFrame
        ds = pd.DataFrame(ds)

        # Return dataset
        return ds

    def get_storm_id(self, storm):
        r"""
        Returns the storm ID (e.g., "AL012019") given the storm name and year.

        Parameters
        ----------
        storm : tuple
            Tuple containing the storm name and year (e.g., ("Matthew",2016)).

        Returns
        -------
        str or list
            If a single storm was found, returns a string containing its ID. Otherwise returns a list of matching IDs.
        """

        # Error check
        if not isinstance(storm, tuple):
            raise TypeError("storm must be of type tuple.")
        if len(storm) != 2:
            raise ValueError(
                "storm must contain 2 elements, name (str) and year (int)")
        name, year = storm

        # Search for corresponding entry in keys
        keys_use = []
        for key in self.dict.keys():
            temp_year = self.dict[key]['year']
            if temp_year == year:
                temp_name = self.dict[key]['name']
                if temp_name == name.upper():
                    keys_use.append(key)

        # return key, or list of keys
        if len(keys_use) == 1:
            keys_use = keys_use[0]
        if len(keys_use) == 0:
            raise RuntimeError("Storm not found")
        return keys_use

    def get_storm(self, storm):
        r"""
        Retrieves a Storm object for the requested storm.

        Parameters
        ----------
        storm : str or tuple
            Requested storm. Can be either string of storm ID (e.g., "AL052019"), or tuple with storm name and year (e.g., ("Matthew",2016)).

        Returns
        -------
        tropycal.tracks.Storm
            Object containing information about the requested storm, and methods for analyzing and plotting the storm.
        """

        # Check if storm is str or tuple
        if isinstance(storm, str):
            key = storm
        elif isinstance(storm, tuple):
            key = self.get_storm_id((storm[0], storm[1]))
        else:
            raise RuntimeError(
                "Storm must be a string (e.g., 'AL052019') or tuple (e.g., ('Matthew',2016)).")

        # Retrieve key of given storm
        if isinstance(key, str):
            return Storm(self.dict[key])
        else:
            error_message = ''.join([f"\n{i}" for i in key])
            error_message = f"Multiple IDs were identified for the requested storm. Choose one of the following storm IDs and provide it as the 'storm' argument instead of a tuple:{error_message}"
            raise RuntimeError(error_message)

    def plot(self, domain=None, ax=None, cartopy_proj=None, save_path=None, **kwargs):
        r"""
        Creates a plot of this season.

        Parameters
        ----------
        domain : str
            Domain for the plot. Default is basin-wide. Please refer to :ref:`options-domain` for available domain options.
        ax : axes
            Instance of axes to plot on. If none, one will be generated. Default is none.
        cartopy_proj : ccrs
            Instance of a cartopy projection to use. If none, one will be generated. Default is none.
        save_path : str
            Relative or full path of directory to save the image in. If none, image will not be saved.

        Other Parameters
        ----------------
        prop : dict
            Customization properties of storm track lines. Please refer to :ref:`options-prop` for available options.
        map_prop : dict
            Customization properties of Cartopy map. Please refer to :ref:`options-map-prop` for available options.

        Returns
        -------
        ax
            Instance of axes containing the plot is returned.
        """

        # Retrieve kwargs
        prop = kwargs.pop('prop', {})
        map_prop = kwargs.pop('map_prop', {})

        # Create instance of plot object
        self.plot_obj = TrackPlot()

        if self.basin in ['east_pacific', 'west_pacific', 'south_pacific', 'australia', 'all']:
            self.plot_obj.create_cartopy(
                proj='PlateCarree', central_longitude=180.0)
        else:
            self.plot_obj.create_cartopy(
                proj='PlateCarree', central_longitude=0.0)

        # Plot storm
        plot_ax = self.plot_obj.plot_season(
            self, domain, ax=ax, save_path=save_path, prop=prop, map_prop=map_prop)

        # Return axis
        return plot_ax

    def summary(self):
        r"""
        Generates a summary for this season with various cumulative statistics.

        Returns
        -------
        dict
            Dictionary containing various statistics about this season.
        """

        # Determine if season object has a single or multiple seasons
        multi_season = isinstance(self.year, list)

        # Initialize dict with info about all of year's storms
        if not multi_season:
            summary_dict = {
                'id': [],
                'operational_id': [],
                'name': [],
                'max_wspd': [],
                'min_mslp': [],
                'category': [],
                'ace': [],
            }
        else:
            summary_dict = {
                'id': [[] for i in range(len(self.year))],
                'operational_id': [[] for i in range(len(self.year))],
                'name': [[] for i in range(len(self.year))],
                'max_wspd': [[] for i in range(len(self.year))],
                'min_mslp': [[] for i in range(len(self.year))],
                'category': [[] for i in range(len(self.year))],
                'ace': [[] for i in range(len(self.year))],
                'seasons': self.year + [],
                'season_start': [0 for i in range(len(self.year))],
                'season_end': [0 for i in range(len(self.year))],
                'season_storms': [0 for i in range(len(self.year))],
                'season_named': [0 for i in range(len(self.year))],
                'season_hurricane': [0 for i in range(len(self.year))],
                'season_major': [0 for i in range(len(self.year))],
                'season_ace': [0 for i in range(len(self.year))],
                'season_subtrop_pure': [0 for i in range(len(self.year))],
                'season_subtrop_partial': [0 for i in range(len(self.year))],
            }

        # Iterate over season(s)
        list_seasons = [self.year] if not multi_season else self.year + []
        for season_idx, iter_season in enumerate(list_seasons):

            # Search for corresponding entry in keys
            count_ss_pure = 0
            count_ss_partial = 0
            iterate_id = 1
            for key in self.dict.keys():

                # Skip if using multi-season object and storm is outside of this season
                if multi_season and self.dict[key]['season'] != iter_season:
                    continue

                # Retrieve info about storm, only in this basin
                temp_name = self.dict[key]['name']
                temp_vmax = np.array(self.dict[key]['vmax'])
                temp_mslp = np.array(self.dict[key]['mslp'])
                temp_type = np.array(self.dict[key]['type'])
                temp_time = np.array(self.dict[key]['time'])
                temp_basin = np.array(self.dict[key]['wmo_basin'])
                temp_year = np.array([i.year for i in self.dict[key]['time']])

                # Calculate ACE within basin
                temp_ace = 0.0
                for ace_i, (i_time, i_vmax, i_basin, i_type) in enumerate(zip(temp_time, temp_vmax, temp_basin, temp_type)):
                    if self.basin not in ['all', 'both'] and i_basin != self.basin:
                        continue
                    if i_time.strftime('%H%M') not in constants.STANDARD_HOURS:
                        continue
                    if i_type not in constants.NAMED_TROPICAL_STORM_TYPES:
                        continue
                    if self.basin == 'all' and i_time.year != self.year:
                        continue
                    if np.isnan(i_vmax):
                        continue
                    temp_ace += accumulated_cyclone_energy(i_vmax)
                temp_ace = np.round(temp_ace, 1)

                # Get indices of all tropical/subtropical time steps
                if self.basin == 'all':
                    idx = np.where(((temp_type == 'SS') | (temp_type == 'SD') | (temp_type == 'TD') | (temp_type == 'TS') | (
                        temp_type == 'HU') | (temp_type == 'TY') | (temp_type == 'ST')) & (temp_year == self.year))
                elif self.basin == 'both':
                    idx = np.where(((temp_type == 'SS') | (temp_type == 'SD') | (temp_type == 'TD') | (
                        temp_type == 'TS') | (temp_type == 'HU') | (temp_type == 'TY') | (temp_type == 'ST')))
                else:
                    idx = np.where(((temp_type == 'SS') | (temp_type == 'SD') | (temp_type == 'TD') | (temp_type == 'TS') | (
                        temp_type == 'HU') | (temp_type == 'TY') | (temp_type == 'ST')) & (temp_basin == self.basin))

                # Get times during existence of trop/subtrop storms
                if len(idx[0]) == 0:
                    continue
                trop_time = temp_time[idx]

                if not multi_season:
                    if 'season_start' not in summary_dict.keys():
                        summary_dict['season_start'] = trop_time[0]
                    else:
                        if trop_time[0] < summary_dict['season_start']:
                            summary_dict['season_start'] = trop_time[0]
                    if 'season_end' not in summary_dict.keys():
                        summary_dict['season_end'] = trop_time[-1]
                    else:
                        if trop_time[-1] > summary_dict['season_end']:
                            summary_dict['season_end'] = trop_time[-1]
                else:
                    if summary_dict['season_start'][season_idx] == 0:
                        summary_dict['season_start'][season_idx] = trop_time[0]
                    else:
                        if trop_time[0] < summary_dict['season_start'][season_idx]:
                            summary_dict['season_start'][season_idx] = trop_time[0]
                    if summary_dict['season_end'][season_idx] == 0:
                        summary_dict['season_end'][season_idx] = trop_time[-1]
                    else:
                        if trop_time[-1] > summary_dict['season_end'][season_idx]:
                            summary_dict['season_end'][season_idx] = trop_time[-1]

                # Get max/min values and check for nan's
                np_wnd = np.array(temp_vmax[idx])
                np_slp = np.array(temp_mslp[idx])
                if len(np_wnd[~np.isnan(np_wnd)]) == 0:
                    max_wnd = np.nan
                    max_cat = -1
                else:
                    max_wnd = int(np.nanmax(temp_vmax[idx]))
                    max_cat = wind_to_category(np.nanmax(temp_vmax[idx]))
                if len(np_slp[~np.isnan(np_slp)]) == 0:
                    min_slp = np.nan
                else:
                    min_slp = int(np.nanmin(temp_mslp[idx]))

                # Append to dict
                if not multi_season:
                    summary_dict['id'].append(key)
                    summary_dict['name'].append(temp_name)
                    summary_dict['max_wspd'].append(max_wnd)
                    summary_dict['min_mslp'].append(min_slp)
                    summary_dict['category'].append(max_cat)
                    summary_dict['ace'].append(temp_ace)
                    summary_dict['operational_id'].append(
                        self.dict[key]['operational_id'])
                else:
                    summary_dict['id'][season_idx].append(key)
                    summary_dict['name'][season_idx].append(temp_name)
                    summary_dict['max_wspd'][season_idx].append(max_wnd)
                    summary_dict['min_mslp'][season_idx].append(min_slp)
                    summary_dict['category'][season_idx].append(max_cat)
                    summary_dict['ace'][season_idx].append(temp_ace)
                    summary_dict['operational_id'][season_idx].append(
                        self.dict[key]['operational_id'])

                # Handle operational vs. non-operational storms

                # Check for purely subtropical storms
                if 'SS' in temp_type and True not in np.isin(temp_type, list(constants.TROPICAL_ONLY_STORM_TYPES)):
                    count_ss_pure += 1

                # Check for partially subtropical storms
                if 'SS' in temp_type:
                    count_ss_partial += 1

            # Add generic season info
            if not multi_season:
                narray = np.array(summary_dict['max_wspd'])
                narray = narray[~np.isnan(narray)]
                summary_dict['season_storms'] = len(narray[narray >= 0])
                summary_dict['season_named'] = len(narray[narray >= 34])
                summary_dict['season_hurricane'] = len(narray[narray >= 65])
                summary_dict['season_major'] = len(narray[narray >= 100])
                summary_dict['season_ace'] = np.round(
                    np.sum(summary_dict['ace']), 1)
                summary_dict['season_subtrop_pure'] = count_ss_pure
                summary_dict['season_subtrop_partial'] = count_ss_partial
            else:
                narray = np.array(summary_dict['max_wspd'][season_idx])
                narray = narray[~np.isnan(narray)]
                summary_dict['season_storms'][season_idx] = len(
                    narray[narray >= 0])
                summary_dict['season_named'][season_idx] = len(
                    narray[narray >= 34])
                summary_dict['season_hurricane'][season_idx] = len(
                    narray[narray >= 65])
                summary_dict['season_major'][season_idx] = len(
                    narray[narray >= 100])
                summary_dict['season_ace'][season_idx] = np.round(
                    np.sum(summary_dict['ace'][season_idx]), 1)
                summary_dict['season_subtrop_pure'][season_idx] = count_ss_pure
                summary_dict['season_subtrop_partial'][season_idx] = count_ss_partial

        # Return object
        return summary_dict



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

import tropycal

sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath(os.path.join('..', '..')))
sys.path.insert(0, os.path.abspath('../tracks'))
sys.path.insert(0, os.path.abspath('../tornado'))
sys.path.insert(0, os.path.abspath('../recon'))
sys.path.insert(0, os.path.abspath('../plot'))

# -- Project information -----------------------------------------------------

project = 'tropycal'
copyright = '2023, Tropycal Developers'
author = 'Tropycal Developers'

# The short X.Y version
version = ''
# The full version, including alpha/beta/rc tags
release = '1.0.1'


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
needs_sphinx = '2.1'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx_gallery.gen_gallery',
    'matplotlib.sphinxext.plot_directive',
]

sphinx_gallery_conf = {
    'doc_module': ('tropycal',),
    'reference_url': {
        'tropycal': None,
    },
    'examples_dirs': [os.path.join('..', 'examples')],
    'gallery_dirs': ['examples'],
    'filename_pattern': r'\.py',
    'backreferences_dir': os.path.join('api', 'generated'),
    'abort_on_example_error': True
}

# Turn off code and image links for embedded mpl plots
plot_html_show_source_link = False
plot_html_show_formats = False

# Tweak how docs are formatted
napoleon_use_rtype = False
napoleon_include_private_with_doc = False

# Control main class documentation
autoclass_content = 'both'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# Controlling automatically generating summary tables in the docs
autosummary_generate = True
autosummary_imported_members = True

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'plot.py',
                   'tracks/plot.py',
                   'tropycal/tracks/plot.py',
                   'src/tropycal/tracks/plot.py',
                   '../src/tropycal/tracks/plot.py']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

mathjax3_config = {'chtml': {'displayAlign': 'left'}}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
#html_theme = 'alabaster'
import sphinx_rtd_theme

html_theme = 'sphinx_rtd_theme'

html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = ['css/theme.css']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = ' '.join((project, version))

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = os.path.join('_static', 'logo.png')

html_favicon = os.path.join('_static', 'logo_32.ico')

html_context = {
    'doc_path': 'docs',
    'galleries': sphinx_gallery_conf['gallery_dirs'],
    'gallery_dir': dict(zip(sphinx_gallery_conf['gallery_dirs'],
                            sphinx_gallery_conf['examples_dirs'])),
    'api_dir': 'api/generated',
    'github_repo': 'tropycal/tropycal',
    'github_version': 'master',  # Make changes to the master branch
}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'tropycaldoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'tropycal.tex', 'tropycal Documentation',
     'Tropycal Developers', 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'tropycal', 'tropycal Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'tropycal', 'tropycal Documentation',
     author, 'tropycal', 'One line description of project.',
     'Miscellaneous'),
]

#--------------- autosummary -------------------------------------------------

napoleon_custom_sections = [('Other Parameters','params_style')]


"""
============================
Aircraft Recon Data Analysis
============================
This sample script shows how to use Tropycal to retrieve, analyze and plot aircraft reconnaissance missions.
"""

from tropycal import tracks, recon
import numpy as np
import datetime as dt

###########################################
# Reading In Recon Dataset
# ------------------------
# There are several ways of retrieving a ReconDataset object. The first step is to retrieve a TrackDataset object, which for this example we'll use the HURDATv2 database for the North Atlantic basin.

basin = tracks.TrackDataset('north_atlantic')

###########################################
# We'll now retrieve a Storm object for Hurricane Irma (2017), which will be used for this example code.

storm = basin.get_storm(('Irma',2017))

###########################################
# Now that we have a Storm object, there are several ways of retrieving recon data.
#
# 1. The first and easiest way is through the Storm object, which already contains an empty instance of ReconDataset stored as an attribute ``storm.recon``. This method will be highlighted in this script. Its methods can then be accessed as ``storm.recon.plot_summary()``, for example.
#
# 2. The second method is through retrieving an instance of ReconDataset, providing the storm object as an argument (e.g., ``recon_obj = recon.ReconDataset(storm)``). Its methods can then be accessed as ``recon_obj.plot_summary()``, for example.
#
# 3. The individual sub-classes (to be discussed later) can also be created individually of a ReconDataset object, providing the storm object as an argument (e.g., ``dropsondes = recon.dropsondes(storm)``).
#
# For the rest of this script, we'll be using the first method.
#
# -----------------
# Recon Sub-Classes
# -----------------
# The three primary sub-classes of the recon module are:
#
# .. list-table:: 
#    :widths: 25 75
#    :header-rows: 1
#
#    * - Class
#      - Description
#    * - hdobs
#      - Class containing all High Density Observations (HDOBs) for this Storm.
#    * - dropsondes
#      - Class containing all dropsondes for this Storm.
#    * - vdms
#      - Class containing all Vortex Data Messages (VDMs) for this Storm.
#
# The following functionality is used to retrieve data for each sub-class. Each class has a ``to_pickle()`` method, which can be used to save the data once it's been read in as a local pickle file, which can be re-read in later.

storm.recon.get_vdms()
#Save pickle file of VDM data (list of dictionaries)
storm.recon.vdms.to_pickle(f'{storm.name}{storm.year}_vdms.pickle')

storm.recon.get_dropsondes()
#Save pickle file of Dropsonde data (list of dictionaries)
storm.recon.dropsondes.to_pickle(f'{storm.name}{storm.year}_dropsondes.pickle')

storm.recon.get_hdobs()
#Save pickle file of HDOB data (Pandas dataframe)
storm.recon.hdobs.to_pickle(f'{storm.name}{storm.year}_hdobs.pickle')

###########################################
# These can be initialized again any time later with the saved pickle files:
#
# >>> storm.recon.get_vdms(f'{storm.name}{storm.year}_vdms.pickle')
# >>> storm.recon.get_dropsondes(f'{storm.name}{storm.year}_dropsondes.pickle')
# >>> storm.recon.get_hdobs(f'{storm.name}{storm.year}_hdobs.pickle')
#

###########################################
# Visualizing ReconDataset
# ------------------------
# Print the recon object to see a summary of the data in recon from the three objects:

storm.recon

###########################################
# A summary of recon data for this storm can also be plotted:

storm.recon.plot_summary()

###########################################
# Find the mission numbers that were active during a given time, within a distance (in km) from the storm:

storm.recon.find_mission(dt.datetime(2017,9,7,12), distance=200)

###########################################
# Then plot a summary from just that mission:

storm.recon.plot_summary(mission=17)

###########################################
# High Density Observations (HDOBs)
# ---------------------------------
# The first class we'll be reviewing is the HDOBs class, which is the largest containing the most data given the high frequency of observations. Let's start by viewing the HDOB summary:

storm.recon.hdobs

###########################################
# And view HDOB Pandas DataFrame data:

storm.recon.hdobs.data

###########################################
# Plot a summary of the recon data for this storm, using peak wind gusts with a custom colormap:

storm.recon.hdobs.plot_points('pkwnd',prop={'cmap':{1:'dodgerblue',2:'gold',3:'firebrick'},'levels':np.arange(20,161,10)})

###########################################
# Plot a hovmoller from recon data interpolated to time and radius - note the eyewall replacement cycles:

storm.recon.hdobs.plot_hovmoller(varname='pkwnd',prop={'cmap':{1:'dodgerblue',2:'gold',3:'firebrick'},'levels':np.arange(20,161,10)})

###########################################
# Plot a map valid at 1200 UTC 6 September 2017 interpolated to time and space:

time = dt.datetime(2017,9,6,12)
storm.recon.hdobs.plot_maps(time=time,varname='pkwnd',prop={'cmap':{1:'dodgerblue',2:'gold',3:'firebrick'},'levels':np.arange(20,161,10)})

###########################################
# Dropsonde Data
# --------------
# Next we'll take a look at the dropsonde data for Hurricane Irma. First, let's take a look at the dropsonde summary:

storm.recon.dropsondes

###########################################
# Now use the ``sel`` function to subset to only dropsondes released in the eyewall:

storm.recon.dropsondes.sel(location='eyewall')

###########################################
# We can view data from one of the dropsondes by using the ``isel`` method to select a dropsonde number:

storm.recon.dropsondes.sel(location='eyewall').isel(23).data

###########################################
# Select one of the eyewall dropsondes and plot the Skew-T:

storm.recon.dropsondes.sel(location='eyewall').isel(23).plot_skewt()

###########################################
# Plot a map of dropsonde points colored by 850mb temperature:

storm.recon.dropsondes.plot_points('temp',level=850,prop={'cmap':{1:'dodgerblue',2:'gold',3:'firebrick'},'ms':20})

###########################################
# Plot a map of only upper-air dropsondes released at 300mb and above, colored by wind speed at 300mb:

storm.recon.dropsondes.sel(top=(None,300)).plot_points('wspd',level=300,prop={'cmap':{1:'dodgerblue',2:'gold',3:'firebrick'},'ms':20})

###########################################
# Vortex Data Messages (VDMs)
# ---------------------------
# The last class we'll look at is the VDM class. Let's start off by viewing a summary of VDM data for Hurricane Irma:

storm.recon.vdms

###########################################
# Let's look at decoded VDMs for a specific pass:

storm.recon.vdms.isel(10).data

###########################################
#Plot a map of VDM center location, colored by minimum pressure (default):

storm.recon.vdms.plot_points(prop={'cmap':{3:'dodgerblue',2:'gold',1:'firebrick'},'ms':40})



r"""Functionality for reading and analyzing SPC tornado dataset."""

import numpy as np
import xarray as xr
import pandas as pd
from datetime import datetime as dt
from scipy.interpolate import griddata
import warnings

from .plot import RainPlot

from ..utils import *


class RainDataset():

    r"""
    Creates an instance of a RainDataset object containing tropical cyclone rainfall data, courtesy of the Weather Prediction Center (WPC).

    Parameters
    ----------
    data_path : str
        Source to read tropical cyclone rainfall data from. Default is "wpc", which reads from the online Weather Prediction Center (WPC) database. Can change this to a local file.

    Returns
    -------
    RainDataset
        An instance of RainDataset.
    """

    def __init__(self, data_path='wpc'):

        # Start timer
        timer_start = dt.now()
        print(f'--> Starting to read in rainfall data')

        # Read in storm rainfall dataset
        if data_path == 'wpc':
            data_path = "https://www.wpc.ncep.noaa.gov/tropical/rain/CONUS_rainfall_obs_1900-2020.csv"
        self.rain_df = pd.read_csv(data_path)

        # Update user on duration
        print(f'--> Completed reading in rainfall data (%.2f seconds)' %
              (dt.now()-timer_start).total_seconds())

    def get_storm_rainfall(self, storm):
        r"""
        Retrieves all rainfall observations in inches associated with a tropical cyclone.

        Parameters
        ----------
        storm : tropycal.tracks.Storm
            Instance of a Storm object.

        Returns
        -------
        pandas.DataFrame
            Pandas DataFrame object containing rainfall data associated with this tropical cyclone, in inches.
        """

        # Filter dataset to this specific storm
        name_1 = f"{storm.name.title()} {storm.year}"
        name_2 = f"{storm.id} {storm.year}"
        df_storm = self.rain_df.loc[(self.rain_df['Storm'] == name_1) | (
            self.rain_df['Storm'] == name_2)]

        # Drop Storm and Year column, no longer necessary
        df_storm = df_storm.drop(columns=['Storm', 'Year'])

        # Remove NaN entries
        df_storm = df_storm.loc[~np.isnan(df_storm['Lat']) & ~np.isnan(
            df_storm['Lon']) & ~np.isnan(df_storm['Total'])]

        # Check if data is empty
        if len(df_storm) == 0:
            raise RuntimeError("No rainfall data is available for this storm.")

        # Return dataframe
        return df_storm

    def interpolate_to_grid(self, storm, grid_res=0.1, method='linear', return_xarray=False):
        r"""
        Interpolates storm rainfall data to a horizontal grid.

        Interpolation is performed using Scipy's `scipy.interpolate.griddata()` interpolation function.

        Parameters
        ----------
        storm : tropycal.tracks.Storm
            Instance of a Storm object.
        grid_res : int or float
            Horizontal resolution of the desired grid in degrees. Default is 0.1 degrees.
        method : str
            Method for interpolation to pass to scipy's interpolation function. Default is "linear".
        return_xarray : bool
            If True, output is returned as an xarray DataArray with coordinates included. Default is false.

        Returns
        -------
        dict or xarray.DataArray
            If return_xarray is True, an xarray DataArray is returned. Otherwise, a dict including the grid lat, lon and grid values is returned.
        """

        # Check if Storm object contains rainfall data
        try:
            storm.rain
        except:
            storm.rain = self.get_storm_rainfall(storm)

        # Create grid to interpolate observations to
        grid_lon = np.arange(-140, -60+grid_res, grid_res)
        grid_lat = np.arange(20, 50+grid_res, grid_res)

        # Retrieve data for interpolation
        rainfall = storm.rain['Total'].values
        lat = storm.rain['Lat'].values
        lon = storm.rain['Lon'].values

        # Perform the interpolation
        grid = griddata((lat, lon), rainfall,
                        (grid_lat[None, :], grid_lon[:, None]), method=method)
        grid = np.transpose(grid)

        # Return data
        if return_xarray:
            return xr.DataArray(grid, coords=[grid_lat, grid_lon], dims=['lat', 'lon'])
        else:
            return {
                'grid': grid,
                'lat': grid_lat,
                'lon': grid_lon
            }

    def plot_rain_grid(self, storm, grid, levels=None, cmap=None, domain="dynamic", plot_all_dots=False, ax=None, cartopy_proj=None, save_path=None, prop={}, map_prop={}):
        r"""
        Creates a plot of a storm track and its associated rainfall (gridded).

        Parameters
        ----------
        storm : tropycal.tracks.Storm
            Storm object to be plotted.
        grid : dict or xarray.DataArray
            Output from `interpolate_to_grid()` to be plotted. Can also be any dict with "lat", "lon" and "grid" entries, or an xarray DataArray with "lat" and "lon" dimensions.
        levels : list or numpy.ndarray
            List of contour fill levels to plot the grid values, in inches. If none, this is automatically generated.
        cmap : colormap
            Colormap to use for contour filling rainfall. If none, this is automatically generated.
        domain : str
            Domain for the plot. Default is "dynamic". Please refer to :ref:`options-domain` for available domain options.
        plot_all_dots : bool
            Whether to plot dots for all observations along the track. If false, dots will be plotted every 6 hours. Default is false.
        ax : axes
            Instance of axes to plot on. If none, one will be generated. Default is none.
        cartopy_proj : ccrs
            Instance of a cartopy projection to use. If none, one will be generated. Default is none.
        save_path : str
            Relative or full path of directory to save the image in. If none, image will not be saved.

        Other Parameters
        ----------------
        prop : dict
            Customization properties of storm track lines. Please refer to :ref:`options-prop` for available options.
        map_prop : dict
            Customization properties of Cartopy map. Please refer to :ref:`options-map-prop` for available options.

        Returns
        -------
        ax
            Instance of axes containing the plot is returned.
        """

        # Check if Storm object contains rainfall data
        try:
            storm.rain
        except:
            storm.rain = self.get_storm_rainfall(storm)

        # Create instance of plot object
        try:
            self.plot_obj
        except:
            self.plot_obj = RainPlot()

        # Create cartopy projection
        if cartopy_proj is not None:
            self.plot_obj.proj = cartopy_proj
        elif max(storm['lon']) > 150 or min(storm['lon']) < -150:
            self.plot_obj.create_cartopy(
                proj='PlateCarree', central_longitude=180.0)
        else:
            self.plot_obj.create_cartopy(
                proj='PlateCarree', central_longitude=0.0)

        # Plot storm
        plot_ax = self.plot_obj.plot_storm(storm, grid, levels, cmap,
                                           domain, plot_all_dots, ax=ax,
                                           save_path=save_path, prop=prop, map_prop=map_prop)

        # Return axis
        return plot_ax

    def plot_rain(self, storm, ms=7.5, mec=None, mew=0.5, minimum_threshold=1.0, levels=None, cmap=None, domain="dynamic", plot_all_dots=False, ax=None, cartopy_proj=None, save_path=None, **kwargs):
        r"""
        Creates a plot of a storm track and its associated rainfall (individual dots).

        Parameters
        ----------
        storm : tropycal.tracks.Storm
            Storm object to be plotted.
        ms : float or int
            Marker size for individual rainfall observations. Default is 7.5.
        mec : str or rgb tuple
            Marker edge color for dots. If none (default), none will be colored.
        mew : float or int
            Marker edge width for dots. If mec is specified, the default is 0.5.
        minimum_threshold : float or int
            Minimum threshold (in inches) to plot dots. Default is 1.00 inch.
        levels : list or numpy.ndarray
            List of levels, in inches, corresponding to the colormap used to color fill the observation dots. If none, this is automatically generated.
        cmap : colormap
            Colormap to use for color filling the observation dots. If none, this is automatically generated.
        domain : str
            Domain for the plot. Default is "dynamic". Please refer to :ref:`options-domain` for available domain options.
        plot_all_dots : bool
            Whether to plot dots for all observations along the track. If false, dots will be plotted every 6 hours. Default is false.
        ax : axes
            Instance of axes to plot on. If none, one will be generated. Default is none.
        cartopy_proj : ccrs
            Instance of a cartopy projection to use. If none, one will be generated. Default is none.
        save_path : str
            Relative or full path of directory to save the image in. If none, image will not be saved.

        Other Parameters
        ----------------
        prop : dict
            Customization properties of storm track lines. Please refer to :ref:`options-prop` for available options.
        map_prop : dict
            Customization properties of Cartopy map. Please refer to :ref:`options-map-prop` for available options.

        Returns
        -------
        ax
            Instance of axes containing the plot is returned.
        """

        prop = kwargs.pop('prop', {})
        map_prop = kwargs.pop('map_prop', {})

        # Check if Storm object contains rainfall data
        try:
            storm.rain
        except:
            storm.rain = self.get_storm_rainfall(storm)

        # Create instance of plot object
        try:
            self.plot_obj
        except:
            self.plot_obj = RainPlot()

        # Create cartopy projection
        if cartopy_proj is not None:
            self.plot_obj.proj = cartopy_proj
        elif max(storm['lon']) > 150 or min(storm['lon']) < -150:
            self.plot_obj.create_cartopy(
                proj='PlateCarree', central_longitude=180.0)
        else:
            self.plot_obj.create_cartopy(
                proj='PlateCarree', central_longitude=0.0)

        # Format plot settings
        plot_settings = {
            'ms': ms,
            'minimum_threshold': minimum_threshold,
            'mew': mew,
            'mec': mec
        }

        # Plot storm
        plot_ax = self.plot_obj.plot_storm(storm, plot_settings, levels,
                                           cmap, domain, plot_all_dots,
                                           ax=ax, save_path=save_path,
                                           prop=prop, map_prop=map_prop)

        # Return axis
        return plot_ax



r"""Functionality for managing real-time tropical cyclone data."""

import re
import ssl
import numpy as np
import urllib
import warnings
from datetime import datetime as dt, timedelta

try:
    import shapefile
    import zipfile
    from io import BytesIO
except ImportError:
    warn_message = "Warning: The libraries necessary for online NHC forecast retrieval aren't available (shapefile, gzip, io)."
    warnings.warn(warn_message)

# Internal imports
from .storm import RealtimeStorm
from ..utils import *
from .. import constants
from ..tracks.plot import TrackPlot


class Realtime():

    r"""
    Creates an instance of a Realtime object containing currently active tropical cyclones and invests.

    Realtime objects are used to retrieve RealtimeStorm objects, which are created for every active tropical cyclone globally upon creating an instance of Realtime.

    Data sources are as follows:
    National Hurricane Center (NHC): https://ftp.nhc.noaa.gov/atcf/btk/
    Joint Typhoon Warning Center (JTWC): (read notes for more details)

    Parameters
    ----------
    jtwc : bool, optional
        Flag determining whether to read JTWC data in. If True, specify the JTWC data source using "jtwc_source". Default is False.
    jtwc_source : str, optional
        If jtwc is set to True, this specifies the JTWC data source to read from. Available options are "noaa", "ucar" or "jtwc". Default is "jtwc". Read the notes for more details.
    ssl_certificate : boolean, optional
        If jtwc is set to True, this determines whether to disable SSL certificate when retrieving data from JTWC. Default is True. Use False *ONLY* if True causes an SSL certification error.

    Returns
    -------
    tropycal.realtime.Realtime
        An instance of Realtime.

    Notes
    -----
    During 2021, the multiple sources offering a Best Track archive of raw JTWC tropical cyclone data experienced frequent outages and/or severe slowdowns, hampering the ability to easily retrieve this data. As such, JTWC data has been optional in Realtime objects since v0.2.7. There are three JTWC sources available:

    * **jtwc** - This is currently the default JTWC source if JTWC data is read in. As of June 2022, this source is working, but reading data is somewhat slow (can take up to 2 minutes).
    * **ucar** - As of September 2021, this source is available and fairly quick to read in, but offers a less compherensive storm history than the "jtwc" source. Between July and September 2021, this source did not update any active tropical cyclones outside of NHC's domain. If using this source, check to make sure it is in fact retrieving current global tropical cyclones.
    * **noaa** - This source was inactive from July 2021 through early 2022, and appears to be back online again. The code retains the ability to read in data from this source should it go back offline then return online again.

    .. warning::

        JTWC's SSL certificate appears to have expired sometime in early 2022. If using JTWC data with the ``jtwc=True`` argument, this will result in Realtime functionality crashing by default. To avoid this, add a ``ssl_certificate=False`` argument to both creating an instance of Realtime and to any method that retrieves JTWC forecast.

        Affected functions include ``Realtime.plot_summary()``, ``RealtimeStorm.get_forecast_realtime()``, and ``RealtimeStorm.plot_forecast_realtime()``.

    The following block of code creates an instance of a Realtime() object and stores it in a variable called "realtime_obj":

    .. code-block:: python

        from tropycal import realtime
        realtime_obj = realtime.Realtime()

    With an instance of Realtime created, any of the methods listed below can be accessed via the "realtime_obj" variable. All active storms and invests are stored as attributes of this instance, and can be simply retrieved:

    .. code-block:: python

        #This stores an instance of a RealtimeStorm object for AL012021 in the "storm" variable.
        storm = realtime_obj.AL012021

        #From there, you can use any method of a Storm object:
        storm.plot()

    RealtimeStorm objects contain all the methods of Storm objects, in addition to special methods available only for storms retrieved via a Realtime object. As of Tropycal v0.3, this now includes invests, though Storm and RealtimeStorm objects retrieved for invests are blocked from performing functionality that does not apply to invests operationally (e.g., forecast discussions, official NHC/JTWC track forecasts).

    To check whether a RealtimeStorm object contains an invest, check the "invest" boolean stored in it:

    .. code-block:: python

        if storm.invest == True:
            print("This is an invest!")
        else:
            print("This is not an invest!")

    Realtime objects have several attributes, including the time the object was last updated, which can be accessed in dictionary format as follows:

    >>> realtime_obj.attrs
    """

    def __repr__(self):

        summary = ["<tropycal.realtime.Realtime>"]

        # Add general summary

        # Add dataset summary
        summary.append("Dataset Summary:")
        summary.append(f'{" "*4}Numbers of active storms: {len(self.storms)}')
        summary.append(
            f'{" "*4}Time Updated: {self.time.strftime("%H%M UTC %d %B %Y")}')

        if len(self.storms) > 0:
            summary.append("\nActive Storms:")
            for key in self.storms:
                if not self[key]['invest']:
                    summary.append(f'{" "*4}{key}')

            summary.append("\nActive Invests:")
            for key in self.storms:
                if self[key]['invest']:
                    summary.append(f'{" "*4}{key}')

        return "\n".join(summary)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __init__(self, jtwc=False, jtwc_source='ucar', ssl_certificate=True):

        # Define empty dict to store track data in
        self.data = {}
        self.jtwc = jtwc
        self.jtwc_source = jtwc_source
        self.ssl_certificate = ssl_certificate

        # Time data reading
        start_time = dt.now()
        print("--> Starting to read in current storm data")

        # Read in best track data
        self.__read_btk()
        if jtwc_source not in ['ucar', 'noaa', 'jtwc']:
            msg = "\"jtwc_source\" must be either \"ucar\", \"noaa\", or \"jtwc\"."
            raise ValueError(msg)
        if jtwc:
            self.__read_btk_jtwc(jtwc_source, ssl_certificate)

        # Determine time elapsed
        time_elapsed = dt.now() - start_time
        tsec = str(round(time_elapsed.total_seconds(), 2))
        print(f"--> Completed reading in current storm data ({tsec} seconds)")

        # Remove storms that haven't been active in 18 hours
        all_keys = [k for k in self.data.keys()]
        for key in all_keys:

            # Filter for storm duration
            if len(self.data[key]['time']) == 0:
                del self.data[key]
                continue

            # Get last time
            last_time = self.data[key]['time'][-1]
            current_time = dt.utcnow()

            # Get time difference
            hours_diff = (current_time - last_time).total_seconds() / 3600.0
            if hours_diff >= 15.0 or (self.data[key]['invest'] and hours_diff >= 9.0):
                del self.data[key]
            if hours_diff <= -48.0:
                del self.data[key]

        # Remove invests that have been classified as TCs
        all_keys = [k for k in self.data.keys()]
        for key in all_keys:

            # Only keep invests
            try:
                if not self.data[key]['invest']:
                    continue
            except:
                continue

            # Iterate through all storms
            match = False
            for key_storm in self.data.keys():
                if self.data[key_storm]['invest']:
                    continue

                # Check for overlap in lons
                if self.data[key_storm]['lon'][0] == self.data[key]['lon'][0] and self.data[key_storm]['time'][0] == self.data[key]['time'][0]:
                    match = True

            if match:
                del self.data[key]

        # For each storm remaining, create a Storm object
        if len(self.data) > 0:
            self.__read_nhc_shapefile()

            # Add probability attributes for storms where it's unavailable
            for key in self.data.keys():
                if key[0:2] in ['AL', 'EP'] and 'prob_2day' not in self.data[key].keys():
                    self.data[key]['prob_2day'] = 'N/A'
                    self.data[key]['prob_7day'] = 'N/A'
                    self.data[key]['risk_2day'] = 'N/A'
                    self.data[key]['risk_7day'] = 'N/A'
                if key[0:2] not in ['AL', 'EP']:
                    self.data[key]['prob_2day'] = 'N/A'
                    self.data[key]['prob_7day'] = 'N/A'
                    self.data[key]['risk_2day'] = 'N/A'
                    self.data[key]['risk_7day'] = 'N/A'
                if self.data[key]['type'][-1] in constants.TROPICAL_STORM_TYPES:
                    self.data[key]['prob_2day'] = 'N/A'
                    self.data[key]['prob_7day'] = 'N/A'
                    self.data[key]['risk_2day'] = 'N/A'
                    self.data[key]['risk_7day'] = 'N/A'
                self[key] = RealtimeStorm(self.data[key])

            # Delete data dict while retaining active storm keys
            self.storms = [k for k in self.data.keys()]
            del self.data
        else:

            # Create an empty list signaling no active storms
            self.storms = []
            del self.data

        # Save current time
        self.time = dt.now()

        # Set attributes
        self.attrs = {
            'jtwc': jtwc,
            'jtwc_source': jtwc_source,
            'time': self.time
        }

    def __read_btk(self):
        r"""
        Reads in best track data into the Dataset object.
        """

        # Get current year
        current_year = (dt.now()).year

        # Get list of files in online directory
        use_ftp = False
        try:
            urlpath = urllib.request.urlopen(
                'https://ftp.nhc.noaa.gov/atcf/btk/')
            string = urlpath.read().decode('utf-8')
        except:
            use_ftp = True
            urlpath = urllib.request.urlopen(
                'ftp://ftp.nhc.noaa.gov/atcf/btk/')
            string = urlpath.read().decode('utf-8')

        # Get relevant filenames from directory
        files = []
        search_pattern = f'b[aec][lp][012349][0123456789]{current_year}.dat'

        pattern = re.compile(search_pattern)
        filelist = pattern.findall(string)
        for filename in filelist:
            if filename not in files:
                files.append(filename)

        # For each file, read in file content and add to hurdat dict
        for file in files:

            # Get file ID
            stormid = ((file.split(".dat")[0])[1:]).upper()

            # Check for invest status
            invest_bool = False
            if int(stormid[2]) == 9:
                invest_bool = True

            # Determine basin
            add_basin = 'north_atlantic'
            if stormid[0] == 'C':
                add_basin = 'east_pacific'
            elif stormid[0] == 'E':
                add_basin = 'east_pacific'

            # add empty entry into dict
            self.data[stormid] = {
                'id': stormid,
                'operational_id': stormid,
                'name': '',
                'year': int(stormid[4:8]),
                'season': int(stormid[4:8]),
                'basin': add_basin,
                'source_info': 'NHC Hurricane Database',
                'realtime': True,
                'invest': invest_bool,
                'source_method': "NHC's Automated Tropical Cyclone Forecasting System (ATCF)",
                'source_url': "https://ftp.nhc.noaa.gov/atcf/btk/",
            }
            self.data[stormid]['source'] = 'hurdat'
            self.data[stormid]['jtwc_source'] = 'N/A'

            # add empty lists
            for val in ['time', 'extra_obs', 'special', 'type', 'lat', 'lon', 'vmax', 'mslp', 'wmo_basin']:
                self.data[stormid][val] = []
            self.data[stormid]['ace'] = 0.0

            # Read in file
            if use_ftp:
                url = f"ftp://ftp.nhc.noaa.gov/atcf/btk/{file}"
            else:
                url = f"https://ftp.nhc.noaa.gov/atcf/btk/{file}"
            f = urllib.request.urlopen(url)
            content = f.read()
            content_full = content.decode("utf-8")
            content = content_full.split("\n")
            content = [(i.replace(" ", "")).split(",") for i in content]
            f.close()

            # Check if transition is in keywords for invests
            if invest_bool and 'TRANSITION' in content_full:
                del self.data[stormid]
                continue

            # iterate through file lines
            for line in content:

                if len(line) < 28:
                    continue

                # Get time of obs
                time = dt.strptime(line[2], '%Y%m%d%H')
                if time.strftime('%H%M') not in constants.STANDARD_HOURS:
                    continue

                # Ensure obs aren't being repeated
                if time in self.data[stormid]['time']:
                    continue

                # Get latitude into number
                btk_lat_temp = line[6].split("N")[0]
                btk_lat = np.round(float(btk_lat_temp) * 0.1, 1)

                # Get longitude into number
                if "W" in line[7]:
                    btk_lon_temp = line[7].split("W")[0]
                    btk_lon = float(btk_lon_temp) * -0.1
                elif "E" in line[7]:
                    btk_lon_temp = line[7].split("E")[0]
                    btk_lon = np.round(float(btk_lon_temp) * 0.1, 1)

                # Get other relevant variables
                btk_wind = int(line[8])
                btk_mslp = int(line[9])
                btk_type = line[10]
                name = line[27]

                # Get last tropical time
                if btk_type in constants.TROPICAL_STORM_TYPES:
                    last_tropical_time = time + timedelta(hours=0)

                # Replace with NaNs
                if btk_wind > 250 or btk_wind < 10:
                    btk_wind = np.nan
                if btk_mslp > 1040 or btk_mslp < 800:
                    btk_mslp = np.nan

                # Add extra obs
                self.data[stormid]['extra_obs'].append(0)

                # Append into dict
                self.data[stormid]['time'].append(time)
                self.data[stormid]['special'].append('')
                self.data[stormid]['type'].append(btk_type)
                self.data[stormid]['lat'].append(round(btk_lat, 1))
                self.data[stormid]['lon'].append(round(btk_lon, 1))
                self.data[stormid]['vmax'].append(btk_wind)
                self.data[stormid]['mslp'].append(btk_mslp)

                # Add basin
                origin_basin = add_basin + ''
                if add_basin == 'east_pacific':
                    check_basin = get_basin(
                        self.data[stormid]['lat'][0], self.data[stormid]['lon'][0], add_basin)
                    if check_basin != add_basin:
                        origin_basin = 'north_atlantic'
                self.data[stormid]['wmo_basin'].append(
                    get_basin(btk_lat, btk_lon, origin_basin))

                # Calculate ACE & append to storm total
                if not np.isnan(btk_wind):
                    ace = accumulated_cyclone_energy(btk_wind)
                    if btk_type in constants.NAMED_TROPICAL_STORM_TYPES:
                        self.data[stormid]['ace'] += np.round(ace, 4)

            # Determine storm name for invests
            if invest_bool:

                # Determine letter in front of invest
                add_letter = 'L'
                if stormid[0] == 'C':
                    add_letter = 'C'
                elif stormid[0] == 'E':
                    add_letter = 'E'
                elif stormid[0] == 'W':
                    add_letter = 'W'
                elif stormid[0] == 'I':
                    add_letter = 'I'
                elif stormid[0] == 'S':
                    add_letter = 'S'
                name = stormid[2:4] + add_letter

            # Add storm name
            self.data[stormid]['name'] = name

            # Check if storm is still tropical, if not an invest.
            # Re-designate as an invest if has not been a TC for over 18 hours.
            if any(type in self.data[stormid]['type'] for type in constants.TROPICAL_STORM_TYPES):
                current_time = dt.utcnow()
                hour_diff = (current_time -
                             last_tropical_time).total_seconds() / 3600
                if hour_diff > 18:
                    self.data[stormid]['invest'] = True

    def __read_btk_jtwc(self, source, ssl_certificate):
        r"""
        Reads in b-deck data from the Tropical Cyclone Guidance Project (TCGP) into the Dataset object.
        """

        # Get current year
        current_year = (dt.now()).year

        # Get list of files in online directory
        url = f'https://www.nrlmry.navy.mil/atcf_web/docs/tracks/{current_year}/'
        if source == 'noaa':
            url = f'https://www.ssd.noaa.gov/PS/TROP/DATA/ATCF/JTWC/'
        if source == 'ucar':
            url = f'http://hurricanes.ral.ucar.edu/repository/data/bdecks_open/{current_year}/'
        if not ssl_certificate and source in ['jtwc', 'noaa']:
            urlpath = urllib.request.urlopen(
                url, context=ssl._create_unverified_context())
        else:
            urlpath = urllib.request.urlopen(url)
        string = urlpath.read().decode('utf-8')

        # Get relevant filenames from directory
        files = []
        search_pattern = f'b[isw][ohp][012349][0123456789]{current_year}.dat'

        pattern = re.compile(search_pattern)
        filelist = pattern.findall(string)
        for filename in filelist:
            if filename not in files:
                files.append(filename)

        # Search for following year (for SH storms)
        search_pattern = f'b[isw][ohp][012349][0123456789]{current_year+1}.dat'

        pattern = re.compile(search_pattern)
        filelist = pattern.findall(string)
        for filename in filelist:
            if filename not in files:
                files.append(filename)

        if source in ['jtwc', 'ucar', 'noaa']:
            try:
                if not ssl_certificate and source in ['jtwc', 'noaa']:
                    urlpath_nextyear = urllib.request.urlopen(url.replace(str(current_year), str(
                        current_year+1)), context=ssl._create_unverified_context())
                    string_nextyear = urlpath_nextyear.read().decode('utf-8')
                else:
                    urlpath_nextyear = urllib.request.urlopen(
                        url.replace(str(current_year), str(current_year+1)))
                    string_nextyear = urlpath_nextyear.read().decode('utf-8')

                pattern = re.compile(search_pattern)
                filelist = pattern.findall(string_nextyear)
                for filename in filelist:
                    if filename not in files:
                        files.append(filename)
            except:
                pass

        # For each file, read in file content and add to hurdat dict
        for file in files:

            # Get file ID
            stormid = ((file.split(".dat")[0])[1:]).upper()

            # Check for invest status
            invest_bool = False
            if int(stormid[2]) == 9:
                invest_bool = True

            # Determine basin based on where storm developed
            add_basin = 'west_pacific'
            if stormid[0] == 'I':
                add_basin = 'north_indian'
            elif stormid[0] == 'S':
                add_basin = ''

            # add empty entry into dict
            self.data[stormid] = {
                'id': stormid,
                'operational_id': stormid,
                'name': '',
                'year': int(stormid[4:8]),
                'season': int(stormid[4:8]),
                'basin': add_basin,
                'source_info': 'Joint Typhoon Warning Center',
                'realtime': True,
                'invest': invest_bool,
            }
            self.data[stormid]['source'] = 'jtwc'
            self.data[stormid]['jtwc_source'] = source

            # Add source info
            self.data[stormid]['source_method'] = "JTWC ATCF"
            self.data[stormid][
                'source_url'] = f'https://www.nrlmry.navy.mil/atcf_web/docs/tracks/{current_year}/'
            if source == 'noaa':
                self.data[stormid]['source_method'] = "NOAA SSD"
                self.data[stormid]['source_url'] = f'https://www.ssd.noaa.gov/PS/TROP/DATA/ATCF/JTWC/'
            if source == 'ucar':
                self.data[stormid][
                    'source_method'] = "UCAR's Tropical Cyclone Guidance Project (TCGP)"
                self.data[stormid]['source_url'] = f'http://hurricanes.ral.ucar.edu/repository/data/bdecks_open/'

            # add empty lists
            for val in ['time', 'extra_obs', 'special', 'type', 'lat', 'lon', 'vmax', 'mslp', 'wmo_basin']:
                self.data[stormid][val] = []
            self.data[stormid]['ace'] = 0.0

            # Read in file
            url = f"https://www.nrlmry.navy.mil/atcf_web/docs/tracks/{current_year}/{file}"
            if source == 'noaa':
                url = f"https://www.ssd.noaa.gov/PS/TROP/DATA/ATCF/JTWC/{file}"
            if source == 'ucar':
                url = f"http://hurricanes.ral.ucar.edu/repository/data/bdecks_open/{current_year}/{file}"
            if f"{current_year+1}.dat" in url:
                url = url.replace(str(current_year), str(current_year+1))

            if not ssl_certificate and source in ['jtwc', 'noaa']:
                f = urllib.request.urlopen(
                    url, context=ssl._create_unverified_context())
                content = f.read()
                content = content.decode("utf-8")
                content = content.split("\n")
                content = [(i.replace(" ", "")).split(",") for i in content]
                f.close()
            else:
                f = urllib.request.urlopen(url)
                content = read_url(url)

            # iterate through file lines
            for line in content:

                if len(line) < 28:
                    continue

                # Get time of obs
                time = dt.strptime(line[2], '%Y%m%d%H')
                if time.strftime('%H%M') not in constants.STANDARD_HOURS:
                    continue

                # Ensure obs aren't being repeated
                if time in self.data[stormid]['time']:
                    continue

                # Get latitude into number
                if "N" in line[6]:
                    btk_lat_temp = line[6].split("N")[0]
                    btk_lat = np.round(float(btk_lat_temp) * 0.1, 1)
                elif "S" in line[6]:
                    btk_lat_temp = line[6].split("S")[0]
                    btk_lat = np.round(float(btk_lat_temp) * -0.1, 1)

                # Get longitude into number
                if "W" in line[7]:
                    btk_lon_temp = line[7].split("W")[0]
                    btk_lon = np.round(float(btk_lon_temp) * -0.1, 1)
                elif "E" in line[7]:
                    btk_lon_temp = line[7].split("E")[0]
                    btk_lon = np.round(float(btk_lon_temp) * 0.1, 1)

                # Determine basin if unknown
                if add_basin == '':
                    add_basin = get_basin(btk_lat, btk_lon)
                    self.data[stormid]['basin'] = add_basin

                # Get other relevant variables
                btk_wind = int(line[8])
                btk_mslp = int(line[9])
                if source == 'ucar':
                    btk_type = get_storm_type(btk_wind, False)
                else:
                    btk_type = line[10]
                name = line[27]

                # Replace with NaNs
                if btk_wind > 250 or btk_wind < 10:
                    btk_wind = np.nan
                if btk_mslp > 1040 or btk_mslp < 800:
                    btk_mslp = np.nan

                # Add extra obs
                self.data[stormid]['extra_obs'].append(0)

                # Append into dict
                self.data[stormid]['time'].append(time)
                self.data[stormid]['special'].append('')
                self.data[stormid]['type'].append(btk_type)
                self.data[stormid]['lat'].append(btk_lat)
                self.data[stormid]['lon'].append(btk_lon)
                self.data[stormid]['vmax'].append(btk_wind)
                self.data[stormid]['mslp'].append(btk_mslp)

                # Add basin
                self.data[stormid]['wmo_basin'].append(
                    get_basin(btk_lat, btk_lon, add_basin))

                # Calculate ACE & append to storm total
                if not np.isnan(btk_wind):
                    ace = accumulated_cyclone_energy(btk_wind)
                    if btk_type in constants.NAMED_TROPICAL_STORM_TYPES:
                        self.data[stormid]['ace'] += np.round(ace, 4)

            # Determine storm name for invests
            if invest_bool:

                # Determine letter in front of invest
                add_letter = 'L'
                if stormid[0] == 'C':
                    add_letter = 'C'
                elif stormid[0] == 'E':
                    add_letter = 'E'
                elif stormid[0] == 'W':
                    add_letter = 'W'
                elif stormid[0] == 'I':
                    add_letter = 'I'
                elif stormid[0] == 'S':
                    add_letter = 'S'
                name = stormid[2:4] + add_letter

            # Add storm name
            self.data[stormid]['name'] = name

    def __read_nhc_shapefile(self):

        try:

            # Read in shapefile zip from NHC
            url = 'https://www.nhc.noaa.gov/xgtwo/gtwo_shapefiles.zip'
            request = urllib.request.Request(url)
            response = urllib.request.urlopen(request)
            file_like_object = BytesIO(response.read())
            tar = zipfile.ZipFile(file_like_object)

            # Get file list (points, areas)
            members = '\n'.join([i for i in tar.namelist()])
            nums = "[0123456789]"
            search_pattern = f'gtwo_points_202{nums}{nums}{nums}{nums}{nums}{nums}{nums}{nums}{nums}.shp'
            pattern = re.compile(search_pattern)
            filelist = pattern.findall(members)
            files = []
            for file in filelist:
                if file not in files:
                    files.append(file.split(".shp")[0])  # remove duplicates

            # Retrieve necessary components for shapefile
            members = tar.namelist()
            members_names = [i for i in members]
            data = {
                'shp': 0,
                'dbf': 0,
                'prj': 0,
                'shx': 0,
            }
            for key in data.keys():
                idx = members_names.index(files[0]+"."+key)
                data[key] = BytesIO(tar.read(members[idx]))

            # Read in shapefile
            orig_reader = shapefile.Reader(
                shp=data['shp'], dbf=data['dbf'], prj=data['prj'], shx=data['shx'])
            shp = BasicReader(orig_reader)

            # Iterate through all areas to match to existing invests
            for record, point in zip(shp.records(), shp.geometries()):

                # Read relevant data
                lon = (list(point.coords)[0][0])
                lat = (list(point.coords)[0][1])
                prob_2day = record.attributes['PROB2DAY']
                prob_7day = record.attributes['PROB7DAY']
                risk_2day = record.attributes['RISK2DAY']
                risk_7day = record.attributes['RISK7DAY']

                # Match to existing invests
                distances = [great_circle((lat, lon), (self.data[storm_id]['lat'][-1],
                                          self.data[storm_id]['lon'][-1])).miles for storm_id in self.data.keys()]
                min_distance = np.min(distances)
                idx = distances.index(min_distance)
                storm_id = [k for k in self.data.keys()][idx]
                if min_distance <= 400:
                    self.data[storm_id]['prob_2day'] = prob_2day
                    self.data[storm_id]['prob_7day'] = prob_7day
                    self.data[storm_id]['risk_2day'] = risk_2day
                    self.data[storm_id]['risk_7day'] = risk_7day

        except:

            msg = "Error in retrieving NHC invest data."
            warnings.warn(msg)

    def update(self):
        r"""
        Update with the latest realtime data.

        Notes
        -----
        This function has no return value, but simply updates the Realtime object with the latest data.
        """

        self.__init__(self.jtwc, self.jtwc_source, self.ssl_certificate)

    def list_active_storms(self, basin='all'):
        r"""
        Produces a list of storms currently stored in Realtime.

        Parameters
        ----------
        basin : str
            Basin for which to return active storms for. Default is 'all'.

        Returns
        -------
        list
            List containing the storm IDs for currently active storms in the requested basin. Each ID has a Storm object stored as an attribute of Realtime.
        """

        if basin == 'all':
            return self.storms

        keys = []
        for key in self.storms:
            if self[key].basin == basin:
                keys.append(key)

        return keys

    def get_storm(self, storm):
        r"""
        Returns a RealtimeStorm object for the requested storm ID.

        Parameters
        ----------
        storm : str
            Storm ID for the requested storm (e.g., "AL012020").

        Returns
        -------
        tropycal.realtime.RealtimeStorm
            An instance of RealtimeStorm.
        """

        # Check to see if storm is available
        if not isinstance(storm, str):
            msg = "\"storm\" must be of type str."
            raise TypeError(msg)
        if storm not in self.storms:
            msg = "Requested storm ID is not contained in this object."
            raise RuntimeError(msg)

        # Return RealtimeStorm object
        return self[storm]

    def plot_summary(self, domain='all', ax=None, cartopy_proj=None, save_path=None, ssl_certificate=True, **kwargs):
        r"""
        Plot a summary map of ongoing tropical cyclone and potential development activity.

        Parameters
        ----------
        domain : str
            Domain for the plot. Default is "all". Please refer to :ref:`options-domain` for available domain options.
        ax : axes, optional
            Instance of axes to plot on. If none, one will be generated. Default is none.
        cartopy_proj : ccrs, optional
            Instance of a cartopy projection to use. If none, one will be generated. Default is none.
        save_path : str, optional
            Relative or full path of directory to save the image in. If none, image will not be saved.
        ssl_certificate : boolean, optional
            If a JTWC forecast, this determines whether to disable SSL certificate when retrieving data from JTWC. Default is True. Use False *ONLY* if True causes an SSL certification error.

        Other Parameters
        ----------------
        two_prop : dict
            Customization properties of NHC Tropical Weather Outlook (TWO). Please refer to :ref:`options-summary` for available options.
        invest_prop : dict
            Customization properties of active invests. Please refer to :ref:`options-summary` for available options.
        storm_prop : dict
            Customization properties of active storms. Please refer to :ref:`options-summary` for available options.
        cone_prop : dict
            Customization properties of cone of uncertainty. Please refer to :ref:`options-summary` for available options.
        map_prop : dict
            Customization properties of Cartopy map. Please refer to :ref:`options-map-prop` for available options.

        Returns
        -------
        ax
            Instance of axes containing the plot is returned.

        Notes
        -----

        The following properties are available for plotting NHC Tropical Weather Outlook (TWO), via ``two_prop``.

        .. list-table:: 
           :widths: 25 75
           :header-rows: 1

           * - Property
             - Description
           * - plot
             - Boolean to determine whether to plot NHC TWO. Default is True.
           * - days
             - Number of days for TWO. Can be either 2 or 5. Default is 5.
           * - fontsize
             - Font size for text label. Default is 12.
           * - ms
             - Marker size for area location, if applicable. Default is 15.

        The following properties are available for plotting invests, via ``invest_prop``.

        .. list-table:: 
           :widths: 25 75
           :header-rows: 1

           * - Property
             - Description
           * - plot
             - Boolean to determine whether to plot active invests. Default is True.
           * - linewidth
             - Line width for past track. Default is 0.8. Set to zero to not plot line.
           * - linecolor
             - Line color for past track. Default is black.
           * - linestyle
             - Line style for past track. Default is dotted.
           * - fontsize
             - Font size for invest name label. Default is 12.
           * - ms
             - Marker size for invest location. Default is 14.

        The following properties are available for plotting storms, via ``storm_prop``.

        .. list-table:: 
           :widths: 25 75
           :header-rows: 1

           * - Property
             - Description
           * - plot
             - Boolean to determine whether to plot active storms. Default is True.
           * - linewidth
             - Line width for past track. Default is 0.8. Set to zero to not plot line.
           * - linecolor
             - Line color for past track. Default is black.
           * - linestyle
             - Line style for past track. Default is dotted.
           * - fontsize
             - Font size for storm name label. Default is 12.
           * - fillcolor
             - Fill color for storm location marker. Default is color by SSHWS category ("category").
           * - label_category
             - Boolean for whether to plot SSHWS category on top of storm location marker. Default is True.
           * - ms
             - Marker size for storm location. Default is 14.

        The following properties are available for plotting realtime cone of uncertainty, via ``cone_prop``.

        .. list-table:: 
           :widths: 25 75
           :header-rows: 1

           * - Property
             - Description
           * - plot
             - Boolean to determine whether to plot cone of uncertainty & forecast track for active storms. Default is True.
           * - linewidth
             - Line width for forecast track. Default is 1.5. Set to zero to not plot line.
           * - alpha
             - Opacity for cone of uncertainty. Default is 0.6.
           * - days
             - Number of days for cone of uncertainty, from 2 through 5. Default is 5.
           * - fillcolor
             - Fill color for forecast dots. Default is color by SSHWS category ("category").
           * - label_category
             - Boolean for whether to plot SSHWS category on top of forecast dots. Default is True.
           * - ms
             - Marker size for forecast dots. Default is 12.
        """

        # Retrieve NHC shapefiles for development areas
        shapefiles = get_two_current()

        # Retrieve kwargs
        two_prop = kwargs.pop('two_prop', {})
        invest_prop = kwargs.pop('invest_prop', {})
        storm_prop = kwargs.pop('storm_prop', {})
        cone_prop = kwargs.pop('cone_prop', {})
        map_prop = kwargs.pop('map_prop', {})

        # Create instance of plot object
        try:
            self.plot_obj
        except:
            self.plot_obj = TrackPlot()

        # Create cartopy projection
        if cartopy_proj is not None:
            self.plot_obj.proj = cartopy_proj
        else:
            self.plot_obj.create_cartopy(
                proj='PlateCarree', central_longitude=180.0)  # 0.0

        # Get realtime forecasts
        forecasts = []
        for key in self.storms:
            if not self[key].invest:
                try:
                    forecasts.append(self.get_storm(
                        key).get_forecast_realtime(ssl_certificate))
                except:
                    forecasts.append({})
            else:
                forecasts.append({})
        forecasts = [entry if 'init' in entry.keys() and (dt.utcnow(
        ) - entry['init']).total_seconds() / 3600.0 <= 12 else {} for entry in forecasts]

        # Plot
        ax = self.plot_obj.plot_summary([self.get_storm(key) for key in self.storms], forecasts,
                                        shapefiles, dt.utcnow(), domain, ax, save_path, two_prop, invest_prop, storm_prop, cone_prop, map_prop)

        return ax



r"""Specify tropycal version."""

def get_version():
    r"""Get the latest version of tropycal."""
    
    try:
        from setuptools_scm import get_version
        return get_version(root='..', relative_to=__file__,
                           version_scheme='post-release', local_scheme='dirty-tag')
    except (ImportError, LookupError):
        from pkg_resources import get_distribution
        return get_distribution(__package__).version


r"""Utility functions that are used across modules for colors.

Public utility functions should be added to documentation in the '/docs/_templates/overrides/tropycal.utils.rst' file."""

import numpy as np
import matplotlib.colors as mcolors
import matplotlib as mlib
import warnings

from .generic_utils import *

# ===========================================================================================================
# Public utilities
# These are used internally and have use externally. Add these to documentation.
# ===========================================================================================================


def get_colors_sshws(wind_speed):
    r"""
    Retrieve the default colors for the Saffir-Simpson Hurricane Wind Scale (SSHWS).

    Parameters
    ----------
    wind_speed : int or list
        Sustained wind speed in knots.

    Returns
    -------
    str
        Hex string for the corresponding color.
    """

    # If category string passed, convert to wind
    if isinstance(wind_speed, str):
        wind_speed = category_label_to_wind(wind_speed)

    # Return default SSHWS category color scale
    if wind_speed < 5:
        return '#FFFFFF'
    elif wind_speed < 34:
        return '#8FC2F2'  # '#7DB7ED'
    elif wind_speed < 64:
        return '#3185D3'
    elif wind_speed < 83:
        return '#FFFF00'
    elif wind_speed < 96:
        return '#FF9E00'
    elif wind_speed < 113:
        return '#DD0000'
    elif wind_speed < 137:
        return '#FF00FC'
    else:
        return '#8B0088'


def get_colors_sshws_recon(wind_speed):
    r"""
    Retrieve modified colors for the Saffir-Simpson Hurricane Wind Scale (SSHWS).

    Parameters
    ----------
    wind_speed : int or list
        Sustained wind speed in knots.

    Returns
    -------
    str
        Hex string for the corresponding color.

    Notes
    -----
    These are the same colors as the default SSHWS colors, but include a special 50-64 knot category.
    """

    # If category string passed, convert to wind
    if isinstance(wind_speed, str):
        wind_speed = category_label_to_wind(wind_speed)

    # Return default SSHWS category color scale
    if wind_speed < 5:
        return '#FFFFFF'
    elif wind_speed < 34:
        return '#8FC2F2'
    elif wind_speed < 50:
        return '#3185D3'
    elif wind_speed < 64:
        return '#05539b'
    elif wind_speed < 83:
        return '#FFFF00'
    elif wind_speed < 96:
        return '#FF9E00'
    elif wind_speed < 113:
        return '#DD0000'
    elif wind_speed < 137:
        return '#FF00FC'
    else:
        return '#8B0088'


def make_colormap(colors, whiten=0):

    z = np.array(sorted(colors.keys()))
    n = len(z)
    z1 = min(z)
    zn = max(z)
    x0 = (z - z1) / (zn - z1)

    CC = mcolors.ColorConverter()
    R = []
    G = []
    B = []
    for i in range(n):
        Ci = colors[z[i]]
        if type(Ci) == str:
            RGB = CC.to_rgb(Ci)
        else:
            RGB = Ci
        R.append(RGB[0] + (1-RGB[0])*whiten)
        G.append(RGB[1] + (1-RGB[1])*whiten)
        B.append(RGB[2] + (1-RGB[2])*whiten)

    cmap_dict = {}
    cmap_dict['red'] = [(x0[i], R[i], R[i]) for i in range(len(R))]
    cmap_dict['green'] = [(x0[i], G[i], G[i]) for i in range(len(G))]
    cmap_dict['blue'] = [(x0[i], B[i], B[i]) for i in range(len(B))]
    mymap = mcolors.LinearSegmentedColormap('mymap', cmap_dict)

    return mymap


def get_colors_ef(colormap='default'):
    r"""
    Retrieve a list of colors for the Enhanced Fujita (EF) tornado scale.

    Parameters
    ----------
    colormap : str or list
        Matplotlib colormap to use. Default is 'default', which uses Tropycal's default colors for the EF scale. If a list, this list must have 6 colors in order from EF0 to EF5.

    Returns
    -------
    list or cmap
        If used a matplotlib colormap, a cmap is returned, otherwise a list of colors is returned.
    """

    # Matplotlib colormap
    if isinstance(colormap, str) and colormap != 'default':
        try:
            cmap = mlib.cm.get_cmap(colormap)
            norm = mlib.colors.Normalize(vmin=0, vmax=5)
            colors = cmap(norm([0, 1, 2, 3, 4, 5]))
        except:
            # colors = [colormap]*6
            raise ValueError('Colormap not found.')

    # User-passed list of colors
    elif isinstance(colormap, list):
        if len(colormap) == 6:
            colors = colormap
        else:
            raise ValueError(
                'Must pass a list of 6 colors to correspond from EF0 to EF5.')

    # Otherwise, return default colors
    else:
        colors = ['lightsalmon', 'tomato', 'red',
                  'firebrick', 'darkred', 'purple']

    # Return list of colors, or colormap
    return colors


def get_colors_pph(plot_type, colormap, levels=None):
    r"""
    Retrieve a list of colors for Practically Perfect Hindcast (PPH) for tornadoes.

    Parameters
    ----------
    plot_type : str
        Plot type for PPH. Can be "daily" for single-day PPH, or "total" for multi-day PPH.
    colormap : str or list
        Matplotlib colormap to use. Default is 'spc', which uses Tropycal's default colors for the EF scale. If a list, this list must have 6 colors in order from EF0 to EF5.
    levels : list
        List of contour levels. Default is SPC intervals.

    Returns
    -------
    colors : list or cmap
        If used a matplotlib colormap, a cmap is returned, otherwise a list of colors is returned.
    levels : list
        List of contour levels.
    """

    # Insert default levels if none specified
    if levels is None:
        levels = [2, 5, 10, 15, 30, 45, 60, 100]

    # Default SPC colormap
    if colormap == 'spc':

        # Daily plot type
        if plot_type == 'daily':
            levels = [2, 5, 10, 15, 30, 45, 60, 100]
            colors = ['#008B00', '#8B4726', '#FFC800',
                      '#FF0000', '#FF00FF', '#912CEE', '#104E8B']

        # Multi-day plot type
        else:
            warnings.warn('SPC colors only allowed for daily PPH.\n' +
                          'Defaulting to plasma colormap.')
            colormap = 'plasma'

    # User-defined colormap
    if colormap != 'spc':

        # Matplotlib colormap
        if isinstance(colormap, str):
            cmap = mlib.cm.get_cmap(colormap)
            norm = mlib.colors.Normalize(vmin=0, vmax=len(levels)-2)
            colors = cmap(norm(np.arange(len(levels))))

        # User defined list of colors
        elif isinstance(colormap, list):
            colors = colormap

        # If a cmap is passed
        else:
            norm = mlib.colors.Normalize(vmin=0, vmax=len(levels)-2)
            colors = colormap(norm(np.arange(len(levels))))

    # Return colormap and values
    return colors, levels

# ===========================================================================================================
# Private utilities
# These are primarily intended to be used internally. Do not add these to documentation.
# ===========================================================================================================


def rgb_tuple_to_str(rgb_tuple):
    r"""
    Convert an RGB tuple to hex string.

    Parameters
    ----------
    rgb_tuple : tuple
        Tuple ordered in (r,g,b) containing integers from 0 to 255.

    Returns
    -------
    str
        Hex string for RGB value.
    """

    r, g, b = rgb_tuple
    r = int(r)
    g = int(g)
    b = int(b)
    return '#%02x%02x%02x' % (r, g, b)


def get_cmap_levels(varname, colormap, levels, linear=False):
    r"""
    Retrieve a list of colors for Practically Perfect Hindcast (PPH) for tornadoes.

    Parameters
    ----------
    varname : str
        if 'vmax', 'sfmr', or 'fl_to_sfc', then SSHWS category colors can be used.
    colormap : str or list
        Matplotlib colormap to use. Default is 'category', which uses Tropycal's default colors for the SSHWS scale. If a list, a colormap is generated from this list.
    levels : list
        List of contour levels. Default is SPC intervals.
    linear : bool
        If colormap is 'category', determine whether to generate a colorbar using linear category increments (True) or wind increments (False).

    Returns
    -------
    colors : cmap
        Matplotlib colormap object.
    levels : list
        List of contour levels.
    """

    # Default SSHWS colormap
    if colormap in ['category', 'category_recon']:

        # Ensure variable contains some element of surface wind
        if varname in ['vmax', 'sfmr', 'wspd', 'fl_to_sfc']:

            # Generate contour levels
            levels = [category_to_wind(c) for c in range(-1, 6)]+[200]

            # Linear category increments
            if linear:
                colors = [mcolors.to_rgba(get_colors_sshws(lev))
                          for c, lev in enumerate(levels[:-1]) for _ in range(levels[c+1]-levels[c])]

            # Linear wind increments
            else:
                if colormap == 'category':
                    levels = [category_to_wind(c) for c in range(-1, 6)]+[200]
                    colors = [get_colors_sshws(lev) for lev in levels[:-1]]
                else:
                    levels = [category_to_wind(
                        c) for c in range(-1, 1)]+[50]+[category_to_wind(c) for c in range(1, 6)]+[200]
                    colors = [get_colors_sshws_recon(
                        lev) for lev in levels[:-1]]
                cmap = mcolors.ListedColormap(colors)

        # Otherwise, default to plasma colormap
        else:
            warnings.warn(
                'Saffir Simpson category colors are not allowed for this variable.')
            colormap = 'plasma'

    # Other colormap options
    if colormap not in ['category', 'category_recon']:

        # Matplotlib colormap name
        if isinstance(colormap, str):
            cmap = mlib.cm.get_cmap(colormap)

        # User defined list of colors
        elif isinstance(colormap, list):
            cmap = mcolors.ListedColormap(colormap)

        # Dictionary
        elif isinstance(colormap, dict):
            cmap = make_colormap(colormap)

        # ListedColormap
        elif isinstance(colormap, mlib.colors.ListedColormap):
            cmap = colormap

        # Default to plasma
        else:
            cmap = mlib.cm.get_cmap('plasma')

        # Normalize colors relative to levels
        norm = mlib.colors.Normalize(vmin=0, vmax=len(levels)-1)

        # If more than 2 levels were passed, use those for the contour levels
        if len(levels) > 2:
            colors = cmap(norm(np.arange(len(levels)-1)))
            cmap = mcolors.ListedColormap(colors)

        # Otherwise, create a list of colors based on levels
        else:
            colors = cmap(norm(np.linspace(0, 1, 256)))
            cmap = mcolors.LinearSegmentedColormap.from_list(
                'my_colormap', colors)

            y0 = min(levels)
            y1 = max(levels)
            dy = (y1-y0)/8
            scalemag = int(np.log(dy)/np.log(10))
            dy_scaled = dy*10**-scalemag
            dc = min([1, 2, 5, 10], key=lambda x: abs(x-dy_scaled))
            dc = dc*10**scalemag
            c0 = np.ceil(y0/dc)*dc
            c1 = np.floor(y1/dc)*dc
            levels = np.arange(c0, c1+dc, dc)

            if scalemag > 0:
                levels = levels.astype(int)

    # Return colormap and levels
    return cmap, levels



"""
======================
Realtime Functionality
======================
This sample script shows how to use Tropycal to retrieve and plot real time tropical cyclones and potential formation. As this script was written on 24 September 2022, the data below is valid for the time this script was written.
"""

from tropycal import realtime

###########################################
# Reading In Realtime Dataset
# ---------------------------
#
# Let's start with the Realtime dataset by loading it into memory. By default, this reads in Best Track data from the National Hurricane Center (NHC) and filters for all active tropical cyclones and invests over the last 18 hours.
# 
# The default is to not include global storms in the Joint Typhoon Warning Center (JTWC)'s area of responsibility. To read in invests and storms within JTWC's domain, add a ``jtwc=True`` argument, followed by a ``jtwc_source`` argument which can be either "jtwc", "noaa", or "ucar". Read through the documentation to see the limits and pros/cons of each method.
# 
# To keep this demonstration simple, we'll solely focus on the North Atlantic basin.

realtime_obj = realtime.Realtime()

###########################################
# View Current Activity
# ---------------------
# 
# To quickly view the content of the Realtime object, we can simply print it:

realtime_obj

###########################################
# Alternatively, we can also list all active storms, along with an optional filter by basin.

realtime_obj.list_active_storms(basin='north_atlantic')

###########################################
# A new functionality with Tropycal v0.4 is to plot a summary of all ongoing activity, whether globally or by basin. This function can be highly customized, with more details in the documentation.
#
# At the time this script was written, Ian, Gaston and Hermine are active in the Atlantic Ocean, with invest 99L (30% of development) active as well.
#
# The domain option for this plot is set to 'all' by default, plotting the full globe. This can also be a basin (e.g., 'north_atlantic', 'east_pacific'), or a custom domain, as we'll plot below.

realtime_obj.plot_summary(domain={'w':-100,'e':-10,'s':4,'n':60})

###########################################
# Let's look at a few ways to customize this plot. There are four properties available to customize the plot, detailed more thoroughly in the documentation:
#
# - ``two_prop`` - Properties to customize NHC Tropical Weather Outlook (TWO) plotting
# - ``invest_prop`` - Properties to customize invest plotting
# - ``storm_prop`` - Properties to customize tropical cyclone plotting
# - ``cone_prop`` - Properties to customize forecast cone/track plotting
#
# The above plot includes the NHC TWO by default. Plotting the TWO overrides any invests that have a TWO associated with them (invests that are outside of NHC's area of responsibility or don't have a TWO still appear). We can pass the ``'plot':False`` argument to any property dict, which in doing so removes that element from the summary map. Let's test this out by removing the TWO and cone of uncertainty from the plot:

realtime_obj.plot_summary(domain='north_atlantic', two_prop={'plot':False}, cone_prop={'plot':False})

###########################################
# Realtime Storms
# ---------------
# To retrieve a storm from a Realtime object, simply use its ``get_storm()`` method and provide an ID as listed in ``list_active_storms()``:

storm = realtime_obj.get_storm('AL092022')

###########################################
# This now returns a RealtimeStorm object. RealtimeStorm objects inherit the same functionality as Storm objects, but have additional functions unique to realtime storms. Additionally, as these can also be valid for invests, certain functionality that is only available for tropical cyclones (e.g., NHC forecasts or discussions) is unavailable for invests.
#
# Let's view what this RealtimeStorm object contains:

storm

###########################################
# A quick and easy way to check if a storm is an invest is by checking its ``invest`` attribute. This will let you know if you can use the full set of functionality available for tropical cyclones or not.

storm.invest

###########################################
# The next few blocks will overview functions unique to RealtimeStorm objects. We can easily retrieve the latest available forecast dictionary from NHC or JTWC, depending on what area of responsibility the storm is in.
#
# This function also calculates the forecast Accumulated Cyclone Energy (ACE), derived by combining its observed ACE through the current time plus the forecast ACE using linearly interpolated forecast sustained wind.

storm.get_forecast_realtime()

###########################################
# We can also plot it using the ``plot_forecast_realtime()`` method:

storm.plot_forecast_realtime()

###########################################
# Storm forecast discussions can also be retrieved for storms in NHC's area of responsibility, or the Prognostic Reasoning product for JTWC's area of responsibility.

storm.get_discussion_realtime()

###########################################
# Lastly, RealtimeStorms also provide the latest available information using the ``get_realtime_info()`` method.
#
# The default argument is ``source='all'``, which returns the latest available data whether from Best Track or NHC Public Advisories. Other possible values are "public_advisory", which only returns the latest public advisory, or "best_track", which only returns the latest best track data.

storm.get_realtime_info()

###########################################
# Realtime Invests
# ----------------
# Invests are essentially RealtimeStorm objects, but without much of the functionality that comes with tropical cyclones (e.g., official forecast track, forecast discussion, etc.). Let's test this out for invest 99L:

invest = realtime_obj.get_storm('AL992022')
invest

###########################################
# As we can see above, the ``invest`` attribute of this object is True. For invests in NHC's area of responsibility, we can retrieve NHC's probability of formation which is matched to the closest TWO to the invest within a certain distance of the invest.

invest.get_realtime_formation_prob()



import numpy as np
import warnings
from datetime import datetime as dt

from ..plot import Plot
from ..utils import *
from .. import constants

try:
    import cartopy.feature as cfeature
    from cartopy import crs as ccrs
except ImportError:
    warnings.warn(
        "Warning: Cartopy is not installed in your python environment. Plotting functions will not work.")

try:
    import matplotlib as mlib
    import matplotlib.lines as mlines
    import matplotlib.patheffects as path_effects
    import matplotlib.pyplot as plt
    import matplotlib.colors as col

except ImportError:
    warnings.warn(
        "Warning: Matplotlib is not installed in your python environment. Plotting functions will not work.")


class RainPlot(Plot):

    def __init__(self):

        self.use_credit = True

    def plot_storm(self, storm, grid, levels, cmap, domain="dynamic", plot_all_dots=False, ax=None, track_labels=False, save_path=None, prop={}, map_prop={}):
        r"""
        Creates a plot of a single storm track.

        Parameters
        ----------
        storm : str, tuple or dict
            Requested storm. Can be either string of storm ID (e.g., "AL052019"), tuple with storm name and year (e.g., ("Matthew",2016)), or a dict entry.
        domain : str
            Domain for the plot. Can be one of the following:
            "dynamic" - default. Dynamically focuses the domain using the storm track(s) plotted.
            "north_atlantic" - North Atlantic Ocean basin
            "pacific" - East/Central Pacific Ocean basin
            "lonW/lonE/latS/latN" - Custom plot domain
        plot_all_dots : bool
            Whether to plot dots for all observations along the track. If false, dots will be plotted every 6 hours. Default is false.
        ax : axes
            Instance of axes to plot on. If none, one will be generated. Default is none.
        prop : dict
            Property of storm track lines.
        map_prop : dict
            Property of cartopy map.
        """

        # Set default properties
        default_prop = {'dots': True, 'fillcolor': 'category', 'cmap': None,
                        'levels': None, 'linecolor': 'k', 'linewidth': 1.0, 'ms': 7.5}
        default_map_prop = {'res': 'm', 'land_color': '#FBF5EA', 'ocean_color': '#EDFBFF',
                            'linewidth': 0.5, 'linecolor': 'k', 'figsize': (14, 9), 'dpi': 200}

        # Initialize plot
        prop = self.add_prop(prop, default_prop)
        map_prop = self.add_prop(map_prop, default_map_prop)
        self.plot_init(ax, map_prop, plot_geography=False)

        # --------------------------------------------------------------------------------------

        # get resolution corresponding to string in prop
        res = self.check_res(map_prop['res'])

        # fill oceans if specified
        self.ax.set_facecolor(map_prop['ocean_color'])
        ocean_mask = self.ax.add_feature(cfeature.OCEAN.with_scale(
            res), facecolor=map_prop['ocean_color'], edgecolor='face', zorder=3)
        lake_mask = self.ax.add_feature(cfeature.LAKES.with_scale(
            res), facecolor=map_prop['ocean_color'], edgecolor='face', zorder=1)
        continent_mask = self.ax.add_feature(cfeature.LAND.with_scale(
            res), facecolor=map_prop['land_color'], edgecolor='face', zorder=0)

        # draw geography
        states = self.ax.add_feature(cfeature.STATES.with_scale(res), linewidths=map_prop['linewidth'],
                                     linestyle='solid', edgecolor=map_prop['linecolor'], zorder=4)
        countries = self.ax.add_feature(cfeature.BORDERS.with_scale(res), linewidths=map_prop['linewidth'],
                                        linestyle='solid', edgecolor=map_prop['linecolor'], zorder=4)
        coastlines = self.ax.add_feature(cfeature.COASTLINE.with_scale(res), linewidths=map_prop['linewidth'],
                                         linestyle='solid', edgecolor=map_prop['linecolor'], zorder=4)

        # --------------------------------------------------------------------------------------

        # Retrieve grid coordinates and values
        plot_grid = True
        if isinstance(grid, dict):
            if 'lat' in grid.keys():
                grid_lat = grid['lat']
                grid_lon = grid['lon']
                grid_val = grid['grid']
            else:
                ms = grid['ms']
                mec = grid['mec']
                mew = grid['mew']
                minimum_threshold = grid['minimum_threshold']
                plot_grid = False
        else:
            grid_lat = grid.lat.values
            grid_lon = grid.lon.values
            grid_val = grid.values

        # Determine levels and colormap
        if levels is None:
            levels = [1, 2, 3, 4, 6, 8, 10, 12, 16, 20]
        if cmap is None:
            cmap = plt.cm.YlGn
        norm = col.BoundaryNorm(levels, cmap.N)

        # Contour fill grid if requested
        if plot_grid:
            self.ax.contourf(grid_lon, grid_lat, grid, levels,
                             cmap=cmap, norm=norm, zorder=2)

        # Plot dots if requested
        else:

            # Iterate over sorted data (so highest totals show up on top)
            iter_df = storm.rain.sort_values('Total')
            for _, row in iter_df.iterrows():

                # Retrieve rain total and determine color
                rain_value = row['Total']
                if rain_value < minimum_threshold:
                    continue
                color = rgb_tuple_to_str(
                    cmap(norm(rain_value), bytes=True)[:-1])

                # Specify additional kwargs
                ms_kwargs = {}
                if mec is not None:
                    ms_kwargs = {'mec': mec, 'mew': mew}
                self.ax.plot(row['Lon'], row['Lat'], 'o', ms=ms,
                             color=color, **ms_kwargs, transform=ccrs.PlateCarree())

        # Produce colorbar
        cs = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
        cs.set_array([])
        self.fig.colorbar(cs, ax=self.ax)

        # --------------------------------------------------------------------------------------

        # Keep record of lat/lon coordinate extrema
        max_lat = np.nanmax(storm.rain['Lat'].values)
        min_lat = np.nanmin(storm.rain['Lat'].values)
        max_lon = np.nanmax(storm.rain['Lon'].values)
        min_lon = np.nanmin(storm.rain['Lon'].values)

        # Get storm data
        storm_data = storm.dict

        # Retrieve storm data
        lats = storm_data['lat']
        lons = storm_data['lon']
        vmax = storm_data['vmax']
        styp = storm_data['type']
        sdate = storm_data['time']

        # Account for cases crossing dateline
        if self.proj.proj4_params['lon_0'] == 180.0:
            new_lons = np.array(lons)
            new_lons[new_lons < 0] = new_lons[new_lons < 0]+360.0
            lons = new_lons.tolist()

        # Force dynamic_tropical to tropical if an invest
        invest_bool = False
        if 'invest' in storm_data.keys() and storm_data['invest']:
            invest_bool = True
            if domain == 'dynamic_tropical':
                domain = 'dynamic'

        # Add to coordinate extrema
        if domain == 'dynamic_tropical':
            type_array = np.array(storm_data['type'])
            idx = np.where((type_array == 'SD') | (type_array == 'SS') | (type_array == 'TD') | (
                type_array == 'TS') | (type_array == 'HU') | (type_array == 'TY') | (type_array == 'ST'))
            use_lats = (np.array(storm_data['lat'])[idx]).tolist()
            use_lons = (np.array(lons)[idx]).tolist()
        else:
            use_lats = storm_data['lat']
            use_lons = np.copy(lons).tolist()

        # Iterate over storm data to plot
        for i, (i_lat, i_lon, i_vmax, i_mslp, i_time, i_type) in enumerate(zip(storm_data['lat'], lons, storm_data['vmax'], storm_data['mslp'], storm_data['time'], storm_data['type'])):

            # Determine line color, with SSHWS scale used as default
            if prop['linecolor'] == 'category':
                segmented_colors = True
                line_color = get_colors_sshws(np.nan_to_num(i_vmax))

            # Use user-defined colormap if another storm variable
            elif isinstance(prop['linecolor'], str) and prop['linecolor'] in ['vmax', 'mslp']:
                segmented_colors = True
                color_variable = storm_data[prop['linecolor']]
                if prop['levels'] is None:  # Auto-determine color levels if needed
                    prop['levels'] = [
                        np.nanmin(color_variable), np.nanmax(color_variable)]
                cmap, levels = get_cmap_levels(
                    prop['linecolor'], prop['cmap'], prop['levels'])
                line_color = cmap((color_variable-min(levels)) /
                                  (max(levels)-min(levels)))[i]

            # Otherwise go with user input as is
            else:
                segmented_colors = False
                line_color = prop['linecolor']

            # For tropical/subtropical types, color-code if requested
            if i > 0:
                if i_type in constants.TROPICAL_STORM_TYPES:

                    # Plot underlying black and overlying colored line
                    self.ax.plot([lons[i-1], lons[i]], [storm_data['lat'][i-1], storm_data['lat'][i]], '-',
                                 linewidth=prop['linewidth']*1.33, color='k', zorder=5,
                                 transform=ccrs.PlateCarree())
                    self.ax.plot([lons[i-1], lons[i]], [storm_data['lat'][i-1], storm_data['lat'][i]], '-',
                                 linewidth=prop['linewidth'], color=line_color, zorder=6,
                                 transform=ccrs.PlateCarree())

                # For non-tropical types, plot dotted lines
                else:

                    # Restrict line width to 1.5 max
                    line_width = prop['linewidth'] + 0.0
                    if line_width > 1.5:
                        line_width = 1.5

                    # Plot dotted line
                    self.ax.plot([lons[i-1], lons[i]], [storm_data['lat'][i-1], storm_data['lat'][i]], ':',
                                 linewidth=line_width, color=line_color, zorder=6,
                                 transform=ccrs.PlateCarree(),
                                 path_effects=[path_effects.Stroke(linewidth=line_width*1.33, foreground='k'),
                                               path_effects.Normal()])

            # Plot dots if requested
            if prop['dots']:

                # Skip if plot_all_dots is False and not in 0,6,12,18z
                if not plot_all_dots:
                    if i_time.strftime('%H%M') not in ['0000', '0600', '1200', '1800']:
                        continue

                # Determine fill color, with SSHWS scale used as default
                if prop['fillcolor'] == 'category':
                    segmented_colors = True
                    fill_color = get_colors_sshws(np.nan_to_num(i_vmax))

                # Use user-defined colormap if another storm variable
                elif isinstance(prop['fillcolor'], str) and prop['fillcolor'] in ['vmax', 'mslp']:
                    segmented_colors = True
                    color_variable = storm_data[prop['fillcolor']]
                    if prop['levels'] is None:  # Auto-determine color levels if needed
                        prop['levels'] = [
                            np.nanmin(color_variable), np.nanmax(color_variable)]
                    cmap, levels = get_cmap_levels(
                        prop['fillcolor'], prop['cmap'], prop['levels'])
                    fill_color = cmap(
                        (color_variable-min(levels))/(max(levels)-min(levels)))[i]

                # Otherwise go with user input as is
                else:
                    segmented_colors = False
                    fill_color = prop['fillcolor']

                # Determine dot type
                marker_type = '^'
                if i_type in constants.SUBTROPICAL_ONLY_STORM_TYPES:
                    marker_type = 's'
                elif i_type in constants.TROPICAL_ONLY_STORM_TYPES:
                    marker_type = 'o'

                # Plot marker
                self.ax.plot(i_lon, i_lat, marker_type, mfc=fill_color, mec='k', mew=0.5,
                             zorder=7, ms=prop['ms'], transform=ccrs.PlateCarree())

            # Label track dots
            if track_labels in ['valid_utc']:
                if track_labels == 'valid_utc':
                    strformat = '%H UTC \n%-m/%-d'
                    labels = {t.strftime(strformat): (x, y) for t, x, y in zip(
                        sdate, lons, lats) if t.hour == 0}
                    track = {t.strftime(strformat): (x, y)
                             for t, x, y in zip(sdate, lons, lats)}
                self.plot_track_labels(self.ax, labels, track, k=.9)

        # --------------------------------------------------------------------------------------

        # Storm-centered plot domain
        if domain == "dynamic" or domain == "dynamic_tropical":

            bound_w, bound_e, bound_s, bound_n = self.dynamic_map_extent(
                min_lon, max_lon, min_lat, max_lat)
            self.ax.set_extent(
                [bound_w, bound_e, bound_s, bound_n], crs=ccrs.PlateCarree())

        # Pre-generated or custom domain
        else:
            bound_w, bound_e, bound_s, bound_n = self.set_projection(domain)

        # Plot parallels and meridians
        # This is currently not supported for all cartopy projections.
        try:
            self.plot_lat_lon_lines(
                [bound_w, bound_e, bound_s, bound_n], zorder=9)
        except:
            pass

        # --------------------------------------------------------------------------------------

        # Add left title
        type_array = np.array(storm_data['type'])
        if not invest_bool:
            idx = np.where((type_array == 'SD') | (type_array == 'SS') | (type_array == 'TD') | (
                type_array == 'TS') | (type_array == 'HU') | (type_array == 'TY') | (type_array == 'ST'))
            tropical_vmax = np.array(storm_data['vmax'])[idx]

            # Coerce to include non-TC points if storm hasn't been designated yet
            add_ptc_flag = False
            if len(tropical_vmax) == 0:
                add_ptc_flag = True
                idx = np.where((type_array == 'LO') | (type_array == 'DB'))
            tropical_vmax = np.array(storm_data['vmax'])[idx]

            subtrop = classify_subtropical(np.array(storm_data['type']))
            peak_idx = storm_data['vmax'].index(np.nanmax(tropical_vmax))
            peak_basin = storm_data['wmo_basin'][peak_idx]
            storm_type = get_storm_classification(
                np.nanmax(tropical_vmax), subtrop, peak_basin)
            if add_ptc_flag:
                storm_type = "Potential Tropical Cyclone"
            left_title_string = f"{storm_type} {storm_data['name']}"
        else:
            # Use all indices for invests
            idx = np.array([True for i in type_array])
            add_ptc_flag = False
            tropical_vmax = np.array(storm_data['vmax'])

            # Determine letter in front of invest
            add_letter = 'L'
            if storm_data['id'][0] == 'C':
                add_letter = 'C'
            elif storm_data['id'][0] == 'E':
                add_letter = 'E'

            # Add title
            left_title_string = f"INVEST {storm_data['id'][2:4]}{add_letter}"

        # Add left title
        if plot_grid:
            left_title_string += "\nInterpolated WPC Storm Rainfall (in)"
        else:
            if minimum_threshold > 1:
                left_title_string += f"\nWPC Storm Rainfall (>{np.round(minimum_threshold,2)} inch)"
            else:
                left_title_string += f"\nWPC Storm Rainfall (inch)"
        self.ax.set_title(left_title_string, loc='left',
                          fontsize=16, fontweight='bold')

        # Add right title
        ace = storm_data['ace']
        if add_ptc_flag:
            ace = 0.0
        type_array = np.array(storm_data['type'])

        # Get storm extrema for display
        mslp_key = 'mslp' if 'wmo_mslp' not in storm_data.keys() else 'wmo_mslp'
        if all_nan(np.array(storm_data[mslp_key])[idx]):
            min_pres = "N/A"
        else:
            min_pres = int(np.nan_to_num(
                np.nanmin(np.array(storm_data[mslp_key])[idx])))
        if all_nan(np.array(storm_data['vmax'])[idx]):
            max_wind = "N/A"
        else:
            max_wind = int(np.nan_to_num(
                np.nanmax(np.array(storm_data['vmax'])[idx])))
        start_time = dt.strftime(
            np.array(storm_data['time'])[idx][0], '%d %b %Y')
        end_time = dt.strftime(np.array(storm_data['time'])[
                               idx][-1], '%d %b %Y')
        endash = u"\u2013"
        dot = u"\u2022"
        self.ax.set_title(
            f"{start_time} {endash} {end_time}\n{max_wind} kt {dot} {min_pres} hPa {dot} {ace:.1f} ACE", loc='right', fontsize=13)

        # --------------------------------------------------------------------------------------

        # Add plot credit
        warning_text = ""
        if storm_data['source'] == 'ibtracs' and storm_data['source_info'] == 'World Meteorological Organization (official)':
            warning_text = f"This plot uses 10-minute averaged WMO official wind data converted\nto 1-minute average (factor of 0.88). Use this wind data with caution.\n\n"

            self.ax.text(0.99, 0.01, warning_text, fontsize=9, color='k', alpha=0.7,
                         transform=self.ax.transAxes, ha='right', va='bottom', zorder=10)

        credit_text = "Rainfall data (inch) from\nWeather Prediction Center (WPC)\n\n"
        self.ax.text(0.99, 0.01, credit_text, fontsize=9, color='k', alpha=0.7,
                     transform=self.ax.transAxes, ha='right', va='bottom', zorder=10)

        credit_text = self.plot_credit()
        self.add_credit(credit_text)

        # --------------------------------------------------------------------------------------

        # Add legend
        if prop['fillcolor'] == 'category' and prop['dots']:
            ex = mlines.Line2D([], [], linestyle='None', ms=prop['ms'],
                               mec='k', mew=0.5, label='Non-Tropical', marker='^', color='w')
            sb = mlines.Line2D([], [], linestyle='None', ms=prop['ms'],
                               mec='k', mew=0.5, label='Subtropical', marker='s', color='w')
            td = mlines.Line2D([], [], linestyle='None', ms=prop['ms'], mec='k', mew=0.5,
                               label='Tropical Depression', marker='o', color=get_colors_sshws(33))
            ts = mlines.Line2D([], [], linestyle='None', ms=prop['ms'], mec='k',
                               mew=0.5, label='Tropical Storm', marker='o', color=get_colors_sshws(34))
            c1 = mlines.Line2D([], [], linestyle='None', ms=prop['ms'], mec='k',
                               mew=0.5, label='Category 1', marker='o', color=get_colors_sshws(64))
            c2 = mlines.Line2D([], [], linestyle='None', ms=prop['ms'], mec='k',
                               mew=0.5, label='Category 2', marker='o', color=get_colors_sshws(83))
            c3 = mlines.Line2D([], [], linestyle='None', ms=prop['ms'], mec='k',
                               mew=0.5, label='Category 3', marker='o', color=get_colors_sshws(96))
            c4 = mlines.Line2D([], [], linestyle='None', ms=prop['ms'], mec='k',
                               mew=0.5, label='Category 4', marker='o', color=get_colors_sshws(113))
            c5 = mlines.Line2D([], [], linestyle='None', ms=prop['ms'], mec='k',
                               mew=0.5, label='Category 5', marker='o', color=get_colors_sshws(137))
            self.ax.legend(handles=[ex, sb, td, ts, c1,
                           c2, c3, c4, c5], prop={'size': 11.5})

        elif prop['linecolor'] == 'category' and not prop['dots']:
            ex = mlines.Line2D([], [], linestyle='dotted',
                               label='Non-Tropical', color='k')
            td = mlines.Line2D([], [], linestyle='solid',
                               label='Sub/Tropical Depression', color=get_colors_sshws(33))
            ts = mlines.Line2D([], [], linestyle='solid',
                               label='Sub/Tropical Storm', color=get_colors_sshws(34))
            c1 = mlines.Line2D([], [], linestyle='solid',
                               label='Category 1', color=get_colors_sshws(64))
            c2 = mlines.Line2D([], [], linestyle='solid',
                               label='Category 2', color=get_colors_sshws(83))
            c3 = mlines.Line2D([], [], linestyle='solid',
                               label='Category 3', color=get_colors_sshws(96))
            c4 = mlines.Line2D([], [], linestyle='solid',
                               label='Category 4', color=get_colors_sshws(113))
            c5 = mlines.Line2D([], [], linestyle='solid',
                               label='Category 5', color=get_colors_sshws(137))
            self.ax.legend(handles=[ex, td, ts, c1, c2, c3, c4, c5], prop={
                           'size': 11.5}).set_zorder(10)

        elif prop['dots'] and not segmented_colors:
            ex = mlines.Line2D([], [], linestyle='None', ms=prop['ms'], mec='k',
                               mew=0.5, label='Non-Tropical', marker='^', color=prop['fillcolor'])
            sb = mlines.Line2D([], [], linestyle='None', ms=prop['ms'], mec='k',
                               mew=0.5, label='Subtropical', marker='s', color=prop['fillcolor'])
            td = mlines.Line2D([], [], linestyle='None', ms=prop['ms'], mec='k',
                               mew=0.5, label='Tropical', marker='o', color=prop['fillcolor'])
            handles = [ex, sb, td]
            self.ax.legend(handles=handles, fontsize=11.5).set_zorder(10)

        elif not prop['dots'] and not segmented_colors:
            ex = mlines.Line2D([], [], linestyle='dotted',
                               label='Non-Tropical', color=prop['linecolor'])
            td = mlines.Line2D([], [], linestyle='solid',
                               label='Tropical', color=prop['linecolor'])
            handles = [ex, td]
            self.ax.legend(handles=handles, fontsize=11.5).set_zorder(10)

        elif prop['dots']:
            ex = mlines.Line2D([], [], linestyle='None', ms=prop['ms'],
                               mec='k', mew=0.5, label='Non-Tropical', marker='^', color='w')
            sb = mlines.Line2D([], [], linestyle='None', ms=prop['ms'],
                               mec='k', mew=0.5, label='Subtropical', marker='s', color='w')
            td = mlines.Line2D([], [], linestyle='None', ms=prop['ms'],
                               mec='k', mew=0.5, label='Tropical', marker='o', color='w')
            handles = [ex, sb, td]
            for _ in range(7):
                handles.append(mlines.Line2D(
                    [], [], linestyle='-', label='', lw=0))
            l = self.ax.legend(handles=handles, fontsize=11.5).set_zorder(10)
            plt.draw()

            # Get the bbox
            try:
                bb = l.legendPatch.get_bbox().inverse_transformed(self.fig.transFigure)
            except:
                bb = l.legendPatch.get_bbox().transformed(self.fig.transFigure.inverted())

            # Define colorbar axis
            cax = self.fig.add_axes(
                [bb.x0+0.47*bb.width, bb.y0+.057*bb.height, 0.015, .65*bb.height])
            norm = mlib.colors.Normalize(vmin=min(levels), vmax=max(levels))
            cbmap = mlib.cm.ScalarMappable(norm=norm, cmap=cmap)
            cbar = self.fig.colorbar(cbmap, cax=cax, orientation='vertical',
                                     ticks=levels)

            cax.tick_params(labelsize=11.5)
            cax.yaxis.set_ticks_position('left')
            cbar.set_label(prop['fillcolor'], fontsize=11.5, rotation=90)

            rect_offset = 0.0
            if prop['cmap'] == 'category' and prop['fillcolor'] == 'vmax':
                cax.yaxis.set_ticks(np.linspace(
                    min(levels), max(levels), len(levels)))
                cax.yaxis.set_ticklabels(levels)
                cax2 = cax.twinx()
                cax2.yaxis.set_ticks_position('right')
                cax2.yaxis.set_ticks((np.linspace(0, 1, len(levels))[
                                     :-1]+np.linspace(0, 1, len(levels))[1:])*.5)
                cax2.set_yticklabels(
                    ['TD', 'TS', 'Cat-1', 'Cat-2', 'Cat-3', 'Cat-4', 'Cat-5'], fontsize=11.5)
                cax2.tick_params('both', length=0, width=0, which='major')
                cax.yaxis.set_ticks_position('left')
                rect_offset = 0.7
            if prop['fillcolor'] == 'time':
                cax.set_yticklabels(
                    [f'{mdates.num2date(i):%b %-d}' for i in clevs], fontsize=11.5)

        else:
            ex = mlines.Line2D([], [], linestyle='dotted',
                               label='Non-Tropical', color='k')
            td = mlines.Line2D([], [], linestyle='solid',
                               label='Tropical', color='k')
            handles = [ex, td]
            for _ in range(7):
                handles.append(mlines.Line2D(
                    [], [], linestyle='-', label='', lw=0))
            l = self.ax.legend(handles=handles, fontsize=11.5).set_zorder(10)
            plt.draw()

            # Get the bbox
            try:
                bb = l.legendPatch.get_bbox().inverse_transformed(self.fig.transFigure)
            except:
                bb = l.legendPatch.get_bbox().transformed(self.fig.transFigure.inverted())

            # Define colorbar axis
            cax = self.fig.add_axes(
                [bb.x0+0.47*bb.width, bb.y0+.057*bb.height, 0.015, .65*bb.height])
            norm = mlib.colors.Normalize(vmin=min(levels), vmax=max(levels))
            cbmap = mlib.cm.ScalarMappable(norm=norm, cmap=cmap)
            cbar = self.fig.colorbar(cbmap, cax=cax, orientation='vertical',
                                     ticks=levels)

            cax.tick_params(labelsize=11.5)
            cax.yaxis.set_ticks_position('left')
            cbarlab = make_var_label(prop['linecolor'], storm_data)
            cbar.set_label(cbarlab, fontsize=11.5, rotation=90)

            rect_offset = 0.0
            if prop['cmap'] == 'category' and prop['linecolor'] == 'vmax':
                cax.yaxis.set_ticks(np.linspace(
                    min(levels), max(levels), len(levels)))
                cax.yaxis.set_ticklabels(levels)
                cax2 = cax.twinx()
                cax2.yaxis.set_ticks_position('right')
                cax2.yaxis.set_ticks((np.linspace(0, 1, len(levels))[
                                     :-1]+np.linspace(0, 1, len(levels))[1:])*.5)
                cax2.set_yticklabels(
                    ['TD', 'TS', 'Cat-1', 'Cat-2', 'Cat-3', 'Cat-4', 'Cat-5'], fontsize=11.5)
                cax2.tick_params('both', length=0, width=0, which='major')
                cax.yaxis.set_ticks_position('left')
                rect_offset = 0.7
            if prop['linecolor'] == 'time':
                cax.set_yticklabels(
                    [f'{mdates.num2date(i):%b %-d}' for i in clevs], fontsize=11.5)

        # -----------------------------------------------------------------------------------------

        # Save image if specified
        if save_path is not None and isinstance(save_path, str):
            plt.savefig(save_path, bbox_inches='tight')

        # Return axis if specified, otherwise display figure
        return self.ax
